# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import itertools
import numpy as np
from numpy.testing import (
    assert_array_equal,
    assert_equal,
)
import pytest
import operator
import warnings

import MDAnalysis as mda
from MDAnalysis.exceptions import NoDataError
from MDAnalysisTests import make_Universe, no_deprecated_call
from MDAnalysisTests.datafiles import PSF, DCD, TPR
from MDAnalysis.core import groups


class TestGroupProperties(object):
    """ Test attributes common to all groups
    """
    @pytest.fixture()
    def u(self):
        return make_Universe(trajectory=True)

    @pytest.fixture()
    def group_dict(self, u):
        return {
            'atom': u.atoms,
            'residue': u.residues,
            'segment': u.segments
        }

    uni = make_Universe() # can't use fixtures in @pytest.mark.parametrize

    def test_dimensions(self, u, group_dict):
        dimensions = np.arange(6)

        for group in group_dict.values():
            group.dimensions = dimensions.copy()
            assert_array_equal(group.dimensions, dimensions)
            assert_equal(u.dimensions, group.dimensions)
    
    @pytest.mark.parametrize('group', (uni.atoms[:2], uni.residues[:2],
                                       uni.segments[:2]))
    def test_group_isunique(self, group):
        assert len(group) == 2
        # Initially, cache must be empty:
        with pytest.raises(KeyError):
            _ = group._cache['isunique']
        # Check for correct value and type:
        assert group.isunique is True
        # Check if cache is set correctly:
        assert group._cache['isunique'] is True

        # Add duplicate element to group:
        group += group[0]
        assert len(group) == 3
        # Cache must be reset since the group changed:
        with pytest.raises(KeyError):
            _ = group._cache['isunique']
        # Check for correct value and type:
        assert group.isunique is False
        # Check if cache is set correctly:
        assert group._cache['isunique'] is False

        #Check empty group:
        group = group[[]]
        assert len(group) == 0
        # Cache must be empty:
        with pytest.raises(KeyError):
            _ = group._cache['isunique']
        # Check for correct value and type:
        assert group.isunique is True
        # Check if cache is set correctly:
        assert group._cache['isunique'] is True

    @pytest.mark.parametrize('group', (uni.atoms[:2], uni.residues[:2],
                                       uni.segments[:2]))
    def test_group_unique_nocache(self, group):
        # check unique group:
        assert len(group) == 2
        # assert caches are empty:
        for attr in ('isunique', 'sorted_unique', 'unsorted_unique'):
            assert attr not in group._cache

    @pytest.mark.parametrize('group', (uni.atoms[:2], uni.residues[:2],
                                       uni.segments[:2]))
    def test_create_unique_group_from_unique(self, group):
        unique_group = group.asunique(sorted=True)
        # assert identity and caches
        assert unique_group is group
        assert group._cache['sorted_unique'] is unique_group
        assert group._cache['unsorted_unique'] is unique_group
        assert group._cache['isunique'] is True
        assert group._cache['issorted']  # numpy.bool != bool

        # assert .unique copies
        assert group.unique is not group
        assert group.unique == group

        # add duplicate element to group:
        group += group[0]
        assert len(group) == 3

        # assert caches are cleared since the group changed:
        for attr in ('isunique', 'sorted_unique', 'unsorted_unique'):
            assert attr not in group._cache

        # now not unique
        assert group.isunique is False
        assert not group.issorted
        # assert that group.unique of the non-unique group is not the old one
        assert group.asunique() is not unique_group
        assert group.asunique() == unique_group
        # check if caches have been set correctly:
        assert group._cache['unsorted_unique'] is group.asunique()
        assert group._cache['sorted_unique'] is group.asunique()
        assert group.unique is not group.asunique()

        # check length and type:
        assert len(group.unique) == 2
        assert type(group.unique) is type(group)

        # check if caches of group.sorted_unique have been set correctly:
        assert group.sorted_unique._cache['isunique'] is True
        assert group.sorted_unique._cache['sorted_unique'] is group.sorted_unique
        # assert that repeated access yields the same object (not a copy):
        unique_group = group.sorted_unique
        assert unique_group is group.sorted_unique

    @pytest.mark.parametrize('ugroup', [uni.atoms, uni.residues, uni.segments])
    @pytest.mark.parametrize('ix, unique_ix', [
        ([0, 1], [0, 1]),
        ([4, 3, 3, 1], [1, 3, 4])
    ])
    def test_group_unique_returns_sorted_copy(self, ugroup, ix, unique_ix):
        # is copy
        group = ugroup[ix]
        assert group.unique is not group
        # sorted
        assert_equal(group.unique.ix, unique_ix)

    @pytest.mark.parametrize('ugroup', [uni.atoms, uni.residues, uni.segments])
    @pytest.mark.parametrize('ix, value', [
        ([4, 3, 3, 1], False),
        ([1, 3, 4], True),
        ([2, 2, 2, 4], True),
    ])
    def test_group_issorted(self, ugroup, ix, value):
        assert ugroup[ix].issorted == value

    @pytest.mark.parametrize('ugroup', [uni.atoms, uni.residues, uni.segments])
    @pytest.mark.parametrize('ix, sort, unique_ix, is_same', [
        ([1, 3, 4], True, [1, 3, 4], True),
        ([1, 3, 4], False, [1, 3, 4], True),
        ([4, 3, 1], True, [1, 3, 4], False),
        ([4, 3, 1], False, [4, 3, 1], True),
        ([1, 3, 3, 4], True, [1, 3, 4], False),
        ([1, 3, 3, 4], False, [1, 3, 4], False),
        ([4, 3, 3, 1], True, [1, 3, 4], False),
        ([4, 3, 3, 1], False, [4, 3, 1], False),
    ])
    def test_group_asunique(self, ugroup, ix, sort, unique_ix, is_same):
        group = ugroup[ix]
        unique_group = group.asunique(sorted=sort)
        assert_equal(unique_group.ix, unique_ix)
        if is_same:
            assert unique_group is group

    @pytest.mark.parametrize('ugroup', [uni.atoms, uni.residues, uni.segments])
    def test_group_return_sorted_unsorted_unique(self, ugroup):
        unsorted_unique = ugroup[[1, 3, 4]].asunique(sorted=False)
        assert 'unsorted_unique' in unsorted_unique._cache
        assert 'sorted_unique' not in unsorted_unique._cache
        assert 'issorted' not in unsorted_unique._cache
        assert 'isunique' in unsorted_unique._cache

        sorted_unique = unsorted_unique.asunique(sorted=True)
        assert sorted_unique is unsorted_unique
        assert unsorted_unique._cache['issorted']
        assert unsorted_unique._cache['sorted_unique'] is unsorted_unique

    @pytest.mark.parametrize('ugroup', [uni.atoms, uni.residues, uni.segments])
    def test_group_return_unsorted_sorted_unique(self, ugroup):
        unique = ugroup[[1, 3, 3, 4]]
        sorted_unique = unique.asunique(sorted=True)
        assert unique._cache['sorted_unique'] is sorted_unique
        assert 'unsorted_unique' not in unique._cache

        unsorted_unique = unique.asunique(sorted=False)
        assert unsorted_unique is sorted_unique
        assert unique._cache['unsorted_unique'] is sorted_unique



class TestGroupSlicing(object):
    """All Groups (Atom, Residue, Segment) should slice like a numpy array

    TODO
    ----
    TopologyGroup is technically called group, add this in too!
    """
    u = make_Universe()

    # test universe is 5:1 mapping 3 times
    group_dict = {
        'atom': u.atoms,
        'residue': u.residues,
        'segment': u.segments
    }
    singulars = {
        'atom': groups.Atom,
        'residue': groups.Residue,
        'segment': groups.Segment
    }
    slices = (
        slice(0, 10),
        slice(0, 2),
        slice(1, 3),
        slice(0, 2, 2),
        slice(0, -1),
        slice(5, 1, -1),
        slice(10, 0, -2),
    )
    length = {'atom': 125,
              'residue': 25,
              'segment': 5}

    levels = ('atom', 'residue', 'segment')

    @pytest.fixture(params=levels)
    def level(self, request):
        return request.param

    @pytest.fixture
    def group(self, level):
        return self.group_dict[level]

    @pytest.fixture
    def nparray(self, level):
        return np.arange(self.length[level])

    @pytest.fixture
    def singular(self, level):
        return self.singulars[level]

    def test_n_atoms(self, group):
        assert len(group.atoms) == group.n_atoms

    def test_n_residues(self, group):
        assert len(group.residues) == group.n_residues

    def test_n_segments(self, group):
        assert len(group.segments) == group.n_segments

    def test_len(self, group, level):
        ref = self.length[level]
        assert len(group) == ref

    @pytest.mark.parametrize('func', [list, np.array])
    def test_boolean_slicing(self, group, func):
        # func is the container type that will be used to slice
        group = group[:5]
        sli = func([True, False, False, True, True])
        result = group[sli]
        assert len(result) == 3
        for ref, val in zip(sli, group):
            if ref:
                assert val in result
            else:
                assert val not in result

    def test_indexerror(self, group, level):
        idx = self.length[level]
        with pytest.raises(IndexError):
            group.__getitem__(idx)

    @pytest.mark.parametrize('sl,func', itertools.product((
        slice(0, 10),
        slice(0, 2),
        slice(1, 3),
        slice(0, 2, 2),
        slice(0, -1),
        slice(5, 1, -1),
        slice(10, 0, -2),
    ), [list, lambda x: np.array(x, dtype=np.int64)]))
    def test_slice(self, group, nparray, sl, func):
        """Check that slicing a np array is identical"""
        g2 = group[sl]
        o2 = nparray[sl]

        assert len(g2) == len(o2)
        # Check identity of items in the sliced result
        for o, g in zip(o2, g2):
            if o in nparray:
                assert g in g2
            else:
                assert g not in g2

    @pytest.mark.parametrize('idx', [0, 1, -1, -2])
    def test_integer_getitem(self, group, nparray, idx, singular):
        a = group[idx]
        ref = nparray[idx]

        assert a.ix == ref
        assert isinstance(a, singular)


def _yield_groups(group_dict, singles, levels, groupclasses, repeat):
    for level in levels:
        for groups in itertools.product([group_dict[level], singles[level]],
                                        repeat=repeat):
            yield  list(groups) + [groupclasses[level]]

class TestGroupAddition(object):
    """Tests for combining Group objects

    Contents
    --------
    Addition of Groups should work like list addition
    Addition of Singular objects should make Group
      A + A -> AG
      AG + A -> AG
      A + AG -> AG
      AG + AG -> AG
    Cross level addition (eg AG + RG) raises TypeError
    Sum() should work on an iterable of many same level Components/Groups
    Groups contain items "x in y"
    """
    u = make_Universe()

    levels = ['atom', 'residue', 'segment']
    group_dict = {
        'atom': u.atoms[:5],
        'residue': u.residues[:5],
        'segment': u.segments[:5],
    }
    singles = {
        'atom': u.atoms[0],
        'residue': u.residues[0],
        'segment': u.segments[0],
    }

    groupclasses = {
        'atom': groups.AtomGroup,
        'residue': groups.ResidueGroup,
        'segment': groups.SegmentGroup,
    }
    # TODO: actually use this
    singleclasses = {
        'atom': groups.Atom,
        'residue': groups.Residue,
        'segment': groups.Segment
    }

    @pytest.fixture(params=levels)
    def level(self, request):
        return request.param

    @pytest.fixture
    def group(self, level):
        return self.group_dict[level]

    @pytest.fixture
    def single(self, level):
        return self.singles[level]

    @pytest.fixture
    def two_groups(self, group, single):
        return itertools.product([group, single], repeat=2)

    @pytest.fixture
    def three_groups(self, group, single):
        return itertools.product([group, single], repeat=3)

    @staticmethod
    def itr(x):
        # singular objects don't iterate
        try:
            x[0]
        except TypeError:
            return [x]
        else:
            return x

    @pytest.mark.parametrize(
        'a, b, refclass',
        _yield_groups(group_dict, singles, levels, groupclasses, repeat=2)
    )
    def test_addition(self, a, b, refclass):
        """Combine a and b, check length, returned type and ordering"""
        newgroup = a + b
        reflen = len(self.itr(a)) + len(self.itr(b))
        assert len(newgroup) == reflen
        assert isinstance(newgroup, refclass)
        # Check ordering of created Group
        for x, y in zip(newgroup, itertools.chain(self.itr(a), self.itr(b))):
            assert x == y

    @pytest.mark.parametrize(
        'a, b, c, refclass',
        _yield_groups(group_dict, singles, levels, groupclasses, repeat=3)
    )
    def test_sum(self, a, b, c, refclass):
        # weird hack in radd allows this
        summed = sum([a, b, c])

        assert isinstance(summed, refclass)
        assert_equal(len(summed),
                     len(self.itr(a)) + len(self.itr(b)) + len(self.itr(c)))
        for x, y in zip(summed,
                        itertools.chain(self.itr(a), self.itr(b), self.itr(c))):
            assert x == y

    @pytest.mark.parametrize(
        'a, b, c, refclass',
        _yield_groups(group_dict, singles, levels, groupclasses, repeat=3)
    )
    def test_bad_sum(self, a, b, c, refclass):
        # sum with bad first argument
        with pytest.raises(TypeError):
            sum([10, a, b, c])

    def test_contains(self, group):
        assert group[2] in group

    def test_contains_false(self, group):
        assert not group[3] in group[:2]

    @pytest.mark.parametrize(
        'one_level, other_level',
        [
            (l1, l2)
            for l1, l2
            in itertools.product(levels, repeat=2)
            if l1 != l2
        ]
    )
    def test_contains_wronglevel(self, one_level, other_level):
        group = self.group_dict[one_level]
        group2 = self.group_dict[other_level]
        assert not group[2] in group2

    @pytest.mark.parametrize(
        'a, b',
        [
            (typeA[alevel], typeB[blevel])
            for (typeA, typeB), (alevel, blevel)
            in itertools.product(
                itertools.product([singles, group_dict], repeat=2),
                itertools.permutations(levels, 2)
            )
        ]
    )
    def test_crosslevel(self, a, b):
        with pytest.raises(TypeError):
            a + b


class TestGroupLevelTransition(object):
    """Test moving between different hierarchy levels

    AtomGroup
    ^
    v
    ResidueGroup
    ^
    v
    SegmentGroup

    *group_to_*group tests moves between levels
    _unique tests check that Upshifts only return unique higher level
    _listcomp tests check that Downshifts INCLUDE repeated elements
    """

    @pytest.fixture()
    def u(self):
        return make_Universe()

    def test_atomgroup_to_atomgroup(self, u):
        atm = u.atoms.atoms
        assert len(atm) == 125
        assert isinstance(atm, groups.AtomGroup)
        assert atm is u.atoms

    def test_atomgroup_to_residuegroup(self, u):
        atm = u.atoms
        res = atm.residues
        assert len(res) == 25
        assert isinstance(res, groups.ResidueGroup)
        assert res == u.residues
        assert res is not u.residues
        assert res._cache['isunique'] is True
        assert res._cache['sorted_unique'] is res

    def test_atomgroup_to_segmentgroup(self, u):
        seg = u.atoms.segments
        assert len(seg) == 5
        assert isinstance(seg, groups.SegmentGroup)
        assert seg == u.segments
        assert seg is not u.segments
        assert seg._cache['isunique'] is True
        assert seg._cache['sorted_unique'] is seg

    def test_residuegroup_to_atomgroup(self, u):
        res = u.residues
        atm = res.atoms
        assert len(atm) == 125
        assert isinstance(atm, groups.AtomGroup)
        assert atm == u.atoms
        assert atm is not u.atoms
        # clear res' uniqueness caches:
        if 'sorted_unique' in res._cache.keys():
            del res._cache['sorted_unique']
        if 'isunique' in res._cache.keys():
            del res._cache['isunique']
        atm = res.atoms
        # assert uniqueness caches of atm are empty:
        with pytest.raises(KeyError):
            _ = atm._cache['isunique']
        with pytest.raises(KeyError):
            _ = atm._cache['sorted_unique']
        # populate uniqueness cache of res:
        assert res.isunique
        atm = res.atoms
        # assert uniqueness caches of atm are set:
        assert atm._cache['isunique'] is True
        assert atm._cache['unsorted_unique'] is atm

    def test_residuegroup_to_residuegroup(self, u):
        res = u.residues.residues
        assert len(res) == 25
        assert isinstance(res, groups.ResidueGroup)
        assert res is u.residues

    def test_residuegroup_to_segmentgroup(self, u):
        seg = u.residues.segments
        assert len(seg) == 5
        assert isinstance(seg, groups.SegmentGroup)
        assert seg == u.segments
        assert seg is not u.segments
        assert seg._cache['isunique'] is True
        assert seg._cache['sorted_unique'] is seg

    def test_segmentgroup_to_atomgroup(self, u):
        seg = u.segments
        atm = seg.atoms
        assert len(atm) == 125
        assert isinstance(atm, groups.AtomGroup)
        assert atm == u.atoms
        assert atm is not u.atoms
        # clear seg's uniqueness caches:
        if 'sorted_unique' in seg._cache.keys():
            del seg._cache['sorted_unique']
        if 'isunique' in seg._cache.keys():
            del seg._cache['isunique']
        atm = seg.atoms
        # assert uniqueness caches of atm are empty:
        with pytest.raises(KeyError):
            _ = atm._cache['isunique']
        with pytest.raises(KeyError):
            _ = atm._cache['sorted_unique']
        # populate uniqueness cache of seg:
        assert seg.isunique
        atm = seg.atoms
        # assert uniqueness caches of atm are set:
        assert atm._cache['isunique'] is True
        assert atm._cache['unsorted_unique'] is atm

    def test_segmentgroup_to_residuegroup(self, u):
        seg = u.segments
        res = seg.residues
        assert len(res) == 25
        assert isinstance(res, groups.ResidueGroup)
        assert res == u.residues
        assert res is not u.residues
        # clear seg's uniqueness caches:
        if 'sorted_unique' in seg._cache.keys():
            del seg._cache['sorted_unique']
        if 'isunique' in seg._cache.keys():
            del seg._cache['isunique']
        res = seg.residues
        # assert uniqueness caches of res are empty:
        with pytest.raises(KeyError):
            _ = res._cache['isunique']
        with pytest.raises(KeyError):
            _ = res._cache['sorted_unique']
        # populate uniqueness cache of seg:
        assert seg.isunique
        res = seg.residues
        # assert uniqueness caches of res are set:
        assert res._cache['isunique'] is True
        assert res._cache['unsorted_unique'] is res

    def test_segmentgroup_to_segmentgroup(self, u):
        seg = u.segments.segments
        assert len(seg) == 5
        assert isinstance(seg, groups.SegmentGroup)
        assert seg is u.segments

    def test_atom_to_residue(self, u):
        res = u.atoms[0].residue
        assert isinstance(res, groups.Residue)

    def test_atom_to_segment(self, u):
        seg = u.atoms[0].segment
        assert isinstance(seg, groups.Segment)

    def test_residue_to_atomgroup(self, u):
        ag = u.residues[0].atoms
        assert isinstance(ag, groups.AtomGroup)
        assert len(ag) == 5
        assert ag._cache['isunique'] is True
        assert ag._cache['sorted_unique'] is ag
        del ag._cache['sorted_unique']
        del ag._cache['isunique']
        assert ag.isunique

    def test_residue_to_segment(self, u):
        seg = u.residues[0].segment
        assert isinstance(seg, groups.Segment)

    def test_segment_to_atomgroup(self, u):
        ag = u.segments[0].atoms
        assert isinstance(ag, groups.AtomGroup)
        assert len(ag) == 25
        assert ag._cache['isunique'] is True
        assert ag._cache['sorted_unique'] is ag
        del ag._cache['sorted_unique']
        del ag._cache['isunique']
        assert ag.isunique

    def test_segment_to_residuegroup(self, u):
        rg = u.segments[0].residues
        assert isinstance(rg, groups.ResidueGroup)
        assert len(rg) == 5
        assert rg._cache['isunique'] is True
        assert rg._cache['sorted_unique'] is rg
        del rg._cache['sorted_unique']
        del rg._cache['isunique']
        assert rg.isunique

    def test_atomgroup_to_residuegroup_unique(self, u):
        ag = u.atoms[:5] + u.atoms[10:15] + u.atoms[:5]
        rg = ag.residues
        assert len(rg) == 2
        assert rg._cache['isunique'] is True
        assert rg._cache['sorted_unique'] is rg

    def test_atomgroup_to_segmentgroup_unique(self, u):
        ag = u.atoms[0] + u.atoms[-1] + u.atoms[0]
        sg = ag.segments
        assert len(sg) == 2
        assert sg._cache['isunique'] is True
        assert sg._cache['sorted_unique'] is sg

    def test_residuegroup_to_segmentgroup_unique(self, u):
        rg = u.residues[0] + u.residues[6] + u.residues[1]
        sg = rg.segments
        assert len(sg) == 2
        assert sg._cache['isunique'] is True
        assert sg._cache['sorted_unique'] is sg

    def test_residuegroup_to_atomgroup_listcomp(self, u):
        rg = u.residues[0] + u.residues[0] + u.residues[4]
        ag = rg.atoms
        assert len(ag) == 15
        # assert uniqueness caches of ag are empty:
        with pytest.raises(KeyError):
            _ = ag._cache['isunique']
        with pytest.raises(KeyError):
            _ = ag._cache['sorted_unique']
        # populate uniqueness cache of rg:
        assert not rg.isunique
        ag = rg.atoms
        # ag uniqueness caches are now from residue
        assert not ag._cache['isunique']
        with pytest.raises(KeyError):
            _ = ag._cache['sorted_unique']

    def test_segmentgroup_to_residuegroup_listcomp(self, u):
        sg = u.segments[0] + u.segments[0] + u.segments[1]
        rg = sg.residues
        assert len(rg) == 15
        # assert uniqueness caches of rg are empty:
        with pytest.raises(KeyError):
            _ = rg._cache['isunique']
        with pytest.raises(KeyError):
            _ = rg._cache['sorted_unique']
        # populate uniqueness cache of sg:
        assert not sg.isunique
        rg = sg.residues
        # assert uniqueness caches of rg are now populated
        assert not rg._cache['isunique']
        with pytest.raises(KeyError):
            _ = rg._cache['sorted_unique']

    def test_segmentgroup_to_atomgroup_listcomp(self, u):
        sg = u.segments[0] + u.segments[0] + u.segments[1]
        ag = sg.atoms
        assert len(ag) == 75
        # assert uniqueness caches of ag are empty:
        with pytest.raises(KeyError):
            _ = ag._cache['isunique']
        with pytest.raises(KeyError):
            _ = ag._cache['sorted_unique']
        # populate uniqueness cache of sg:
        assert not sg.isunique
        ag = sg.atoms
        # ag uniqueness caches are now from segment
        assert not ag._cache['isunique']
        with pytest.raises(KeyError):
            _ = ag._cache['sorted_unique']


class TestComponentComparisons(object):
    """Use of operators (< > == != <= >=) with Atom, Residue, and Segment"""
    u = make_Universe()
    levels = [u.atoms, u.residues, u.segments]

    @pytest.fixture(params=levels)
    def abc(self, request):
        level = request.param
        return level[0], level[1], level[2]

    @pytest.fixture
    def a(self, abc):
        return abc[0]

    @pytest.fixture
    def b (self, abc):
        return abc[1]

    @pytest.fixture
    def c(self, abc):
        return abc[2]

    def test_lt(self, a, b, c):
        assert a < b
        assert a < c
        assert not b < a
        assert not a < a

    def test_gt(self, a, b, c):
        assert b > a
        assert c > a
        assert not a > c
        assert not a > a

    def test_ge(self, a, b, c):
        assert b >= a
        assert c >= a
        assert b >= b
        assert not b >= c

    def test_le(self, a, b, c):
        assert b <= c
        assert b <= b
        assert not b <= a

    def test_neq(self, a, b, c):
        assert a != b
        assert not a != a

    def test_eq(self, a, b, c):
        assert a == a
        assert not a == b

    def test_sorting(self, a, b, c):
        assert sorted([b, a, c]) == [a, b, c]

    @pytest.mark.parametrize(
        'x, y',
        itertools.permutations((u.atoms[0], u.residues[0], u.segments[0]), 2)
    )
    def test_crosslevel_cmp(self, x, y):
        with pytest.raises(TypeError):
            operator.lt(x, y)
        with pytest.raises(TypeError):
            operator.le(x, y)
        with pytest.raises(TypeError):
            operator.gt(x, y)
        with pytest.raises(TypeError):
            operator.ge(x, y)

    @pytest.mark.parametrize(
        'x, y',
        itertools.permutations((u.atoms[0], u.residues[0], u.segments[0]), 2)
    )
    def test_crosslevel_eq(self, x, y):
        with pytest.raises(TypeError):
            operator.eq(x, y)

        with pytest.raises(TypeError):
            operator.ne(x, y)


class TestMetaclassMagic(object):
    # tests for the weird voodoo we do with metaclasses
    def test_new_class(self):
        u = make_Universe(trajectory=True)

        # should be able to subclass AtomGroup as normal
        class NewGroup(groups.AtomGroup):
            pass

        ng = NewGroup(np.array([0, 1, 2]), u)

        assert isinstance(ng, NewGroup)

        ag = u.atoms[[0, 1, 2]]

        assert_array_equal(ng.positions, ag.positions)


class TestGroupBy(object):
    # tests for the method 'groupby'
    @pytest.fixture()
    def u(self):
        return make_Universe(('segids', 'charges', 'resids'))

    def test_groupby_float(self, u):
        gb = u.atoms.groupby('charges')

        for ref in [-1.5, -0.5, 0.0, 0.5, 1.5]:
            assert ref in gb
            g = gb[ref]
            assert all(g.charges == ref)
            assert len(g) == 25

    @pytest.mark.parametrize('string', ['segids', b'segids', u'segids'])
    def test_groupby_string(self, u, string):
        gb = u.atoms.groupby(string)

        assert len(gb) == 5
        for ref in ['SegA', 'SegB', 'SegC', 'SegD', 'SegE']:
            assert ref in gb
            g = gb[ref]
            assert all(g.segids == ref)
            assert len(g) == 25

    def test_groupby_int(self, u):
        gb = u.atoms.groupby('resids')

        for g in gb.values():
            assert len(g) == 5

    # tests for multiple attributes as arguments

    def test_groupby_float_string(self, u):
        gb = u.atoms.groupby(['charges', 'segids'])

        for ref in [-1.5, -0.5, 0.0, 0.5, 1.5]:
            for subref in ['SegA','SegB','SegC','SegD','SegE']:
                assert (ref, subref) in gb.keys()
                a = gb[(ref, subref)]
                assert len(a) == 5
                assert all(a.charges == ref)
                assert all(a.segids  == subref)

    def test_groupby_int_float(self, u):
        gb = u.atoms.groupby(['resids', 'charges'])

        uplim=int(len(gb)/5+1)
        for ref in range(1, uplim):
            for subref in [-1.5, -0.5, 0.0, 0.5, 1.5]:
                assert (ref, subref) in gb.keys()
                a = gb[(ref, subref)]
                assert len(a) == 1
                assert all(a.resids == ref)
                assert all(a.charges == subref)

    def test_groupby_string_int(self, u):
        gb = u.atoms.groupby(['segids', 'resids'])

        assert len(gb) == 25
        res = 1
        for ref in ['SegA','SegB','SegC','SegD','SegE']:
            for subref in range(0, 5):
                assert (ref, res) in gb.keys()
                a = gb[(ref, res)]
                assert all(a.segids == ref)
                assert all(a.resids == res)
                res += 1


class TestReprs(object):
    @pytest.fixture()
    def u(self):
        return mda.Universe(PSF, DCD)

    def test_atom_repr(self, u):
        at = u.atoms[0]
        assert repr(at) == '<Atom 1: N of type 56 of resname MET, resid 1 and segid 4AKE>'

    def test_residue_repr(self, u):
        res = u.residues[0]
        assert repr(res) == '<Residue MET, 1>'

    def test_segment_repr(self, u):
        seg = u.segments[0]
        assert repr(seg) == '<Segment 4AKE>'

    def test_atomgroup_repr(self, u):
        ag = u.atoms[:10]
        assert repr(ag) == '<AtomGroup with 10 atoms>'

    def test_atomgroup_str_short(self, u):
        ag = u.atoms[:2]
        assert str(ag) == '<AtomGroup [<Atom 1: N of type 56 of resname MET, resid 1 and segid 4AKE>, <Atom 2: HT1 of type 2 of resname MET, resid 1 and segid 4AKE>]>'

    def test_atomgroup_str_long(self, u):
        ag = u.atoms[:11]
        assert str(ag).startswith('<AtomGroup [<Atom 1: N of type 56 of resname MET,')
        assert '...' in str(ag)
        assert str(ag).endswith(', resid 1 and segid 4AKE>]>')

    def test_residuegroup_repr(self, u):
        rg = u.residues[:10]
        assert repr(rg) == '<ResidueGroup with 10 residues>'

    def test_residuegroup_str_short(self, u):
        rg = u.residues[:2]
        assert str(rg) == '<ResidueGroup [<Residue MET, 1>, <Residue ARG, 2>]>'

    def test_residuegroup_str_long(self, u):
        rg = u.residues[:11]
        assert str(rg).startswith('<ResidueGroup [<Residue MET, 1>,')
        assert '...' in str(rg)
        assert str(rg).endswith(', <Residue ALA, 11>]>')

    def test_segmentgroup_repr(self, u):
        sg = u.segments[:10]
        assert repr(sg) == '<SegmentGroup with 1 segment>'

    def test_segmentgroup_str(self, u):
        sg = u.segments[:10]
        assert str(sg) == '<SegmentGroup [<Segment 4AKE>]>'


def _yield_mix(groups, components):
    indices = list(range(len(components)))
    for left, right in itertools.permutations(indices, 2):
        yield (groups[left], components[right])
        yield (components[left], groups[right])

def _yield_sliced_groups(u, slice_left, slice_right):
    for level in ('atoms', 'residues', 'segments'):
        yield (getattr(u, level)[slice_left], getattr(u, level)[slice_right])

class TestGroupBaseOperators(object):
    u = make_Universe()

    components = (u.atoms[0], u.residues[0], u.segments[0])
    component_groups = (u.atoms, u.residues, u.segments)

    @pytest.fixture(params=('atoms', 'residues', 'segments'))
    def level(self, request):
        return request.param

    @pytest.fixture
    def groups_simple(self, level):
        n_segments = 10
        n_residues = n_segments * 5
        n_atoms = n_residues * 5
        u = make_Universe(size=(n_atoms, n_residues, n_segments))
        #   0123456789
        # a  ****
        # b    *****
        # c    **
        # e      ***
        # d empty
        #
        # None of the group start at 0, nor ends at the end. Each group
        # has a different size. The end of a slice is not the last element.
        # This increase the odds of catching errors.
        a = getattr(u, level)[1:5]
        b = getattr(u, level)[3:8]
        c = getattr(u, level)[3:5]
        d = getattr(u, level)[0:0]
        e = getattr(u, level)[5:8]
        return a, b, c, d, e

    @pytest.fixture
    def groups_duplicated_and_scrambled(self, level):
        # The content of the groups is the same as for make_groups, but the
        # elements can appear several times and their order is scrambled.
        n_segments = 10
        n_residues = n_segments * 5
        n_atoms = n_residues * 5
        u = make_Universe(size=(n_atoms, n_residues, n_segments))
        a = getattr(u, level)[[1, 3, 2, 1, 2, 4, 4]]
        b = getattr(u, level)[[7, 4, 4, 6, 5, 3, 7, 6]]
        c = getattr(u, level)[[4, 4, 3, 4, 3, 3]]
        d = getattr(u, level)[0:0]
        e = getattr(u, level)[[6, 5, 7, 7, 6]]
        return a, b, c, d, e

    @pytest.fixture(params=('simple', 'scrambled'))
    def groups(self, request, groups_simple, groups_duplicated_and_scrambled):
        return {'simple': groups_simple,
                'scrambled': groups_duplicated_and_scrambled}[request.param]

    def test_len(self, groups_simple):
        a, b, c, d, e = groups_simple
        assert_equal(len(a), 4)
        assert_equal(len(b), 5)
        assert_equal(len(c), 2)
        assert_equal(len(d), 0)
        assert_equal(len(e), 3)

    def test_len_duplicated_and_scrambled(self, groups_duplicated_and_scrambled):
        a, b, c, d, e = groups_duplicated_and_scrambled
        assert_equal(len(a), 7)
        assert_equal(len(b), 8)
        assert_equal(len(c), 6)
        assert_equal(len(d), 0)
        assert_equal(len(e), 5)


    def test_equal(self, groups):
        a, b, c, d, e = groups
        assert a == a
        assert a != b
        assert not a == b
        assert not a[0:1] == a[0], \
            'Element should not equal single element group.'

    @pytest.mark.parametrize('group', (u.atoms[:2], u.residues[:2],
                                       u.segments[:2]))
    def test_copy(self, group):
        # make sure uniqueness caches of group are empty:
        with pytest.raises(KeyError):
            _ = group._cache['isunique']
        with pytest.raises(KeyError):
            _ = group._cache['sorted_unique']
        # make a copy:
        cgroup = group.copy()
        # check if cgroup is an identical copy of group:
        assert type(cgroup) is type(group)
        assert cgroup is not group
        assert cgroup == group
        # check if the copied group's uniqueness caches are empty:
        with pytest.raises(KeyError):
            _ = cgroup._cache['isunique']
        with pytest.raises(KeyError):
            _ = cgroup._cache['sorted_unique']
        # populate group's uniqueness caches:
        assert group.isunique
        # make a copy:
        cgroup = group.copy()
        # check if the copied group's uniqueness caches are set correctly:
        assert cgroup._cache['isunique'] is True
        # assert sorted_unique still doesn't exist
        assert 'sorted_unique' not in cgroup._cache
        # add duplicate element to group:
        group += group[0]
        # populate group's uniqueness caches:
        assert not group.isunique
        # make a copy:
        cgroup = group.copy()
        # check if the copied group's uniqueness caches are set correctly:
        assert cgroup._cache['isunique'] is False
        with pytest.raises(KeyError):
            _ = cgroup._cache['sorted_unique']
        # assert that duplicates are preserved:
        assert cgroup == group

    def test_issubset(self, groups):
        a, b, c, d, e = groups
        assert c.issubset(a)
        assert not c.issubset(e)
        assert not a.issubset(c)
        assert d.issubset(a)
        assert not a.issubset(d)

    def test_is_strict_subset(self, groups):
        a, b, c, d, e = groups
        assert c.is_strict_subset(a)
        assert not c.is_strict_subset(e)
        assert not a.is_strict_subset(a)

    def test_issuperset(self, groups):
        a, b, c, d, e = groups
        assert a.issuperset(c)
        assert not e.issuperset(c)
        assert not c.issuperset(a)
        assert a.issuperset(d)
        assert not d.issuperset(a)

    def test_is_strict_superset(self, groups):
        a, b, c, d, e = groups
        assert a.is_strict_superset(c)
        assert not c.is_strict_superset(e)
        assert not a.is_strict_superset(a)

    def test_concatenate(self, groups):
        a, b, c, d, e = groups
        cat_ab = a.concatenate(b)
        assert cat_ab[:len(a)] == a
        assert cat_ab[len(a):] == b

        cat_ba = b.concatenate(a)
        assert cat_ba[:len(b)] == b
        assert cat_ba[len(b):] == a

        cat_aa = a.concatenate(a)
        assert cat_aa[:len(a)] == a
        assert cat_aa[len(a):] == a

        cat_ad = a.concatenate(d)
        assert cat_ad == a

        cat_da = d.concatenate(a)
        assert cat_da == a

    def test_union(self, groups):
        a, b, c, d, e = groups
        union_ab = a.union(b)
        assert union_ab.ix.tolist() == sorted(union_ab.ix)
        assert list(sorted(set(union_ab.ix))) == list(sorted(union_ab.ix))

        assert a.union(b) == b.union(a)
        assert_array_equal(a.union(a).ix, np.arange(1, 5))
        assert a.union(d), np.arange(1, 5)

    def test_intersection(self, groups):
        a, b, c, d, e = groups
        intersect_ab = a.intersection(b)
        assert_array_equal(intersect_ab.ix, np.arange(3, 5))
        assert a.intersection(b) == b.intersection(a)
        assert_equal(len(a.intersection(d)), 0)

    def test_subtract(self, groups):
        a, b, c, d, e = groups
        subtract_ab = a.subtract(b)
        reference = np.array([1, 2, 1, 2])
        # The groups can be "simple" or "scrambled", if they are simple they
        # do not contain repetitions and they are ordered. The reference for
        # the simple groups should follow the same patern and should be ordered
        # and without repetitions.
        if len(a) == len(np.unique(a)):
            reference = np.unique(reference)
        assert_array_equal(subtract_ab.ix, reference)
        subtract_ba = b.subtract(a)
        reference = np.array([7, 6, 5, 7, 6])
        if len(a) == len(np.unique(a)):
            reference = np.unique(reference)
        assert_array_equal(subtract_ba.ix, reference)
        subtract_ad = a.subtract(d)
        assert_equal(subtract_ad, a)
        subtract_ae = a.subtract(e)
        assert_equal(subtract_ae, a)

    def test_difference(self, groups):
        a, b, c, d, e = groups
        difference_ab = a.difference(b)
        assert_array_equal(difference_ab.ix, np.arange(1, 3))

        difference_ba = b.difference(a)
        assert_array_equal(difference_ba.ix, np.arange(5, 8))

        assert_array_equal(a.difference(d).ix, np.arange(1, 5))
        assert_array_equal(a.difference(e).ix, np.arange(1, 5))

    def test_symmetric_difference(self, groups):
        a, b, c, d, e = groups
        symdiff_ab = a.symmetric_difference(b)
        assert_array_equal(symdiff_ab.ix, np.array(list(range(1, 3)) +
                                                   list(range(5, 8))))
        assert a.symmetric_difference(b) == b.symmetric_difference(a)
        assert_array_equal(a.symmetric_difference(e).ix, np.arange(1, 8))

    def test_isdisjoint(self, groups):
        a, b, c, d, e = groups
        assert a.isdisjoint(e)
        assert e.isdisjoint(a)
        assert a.isdisjoint(d)
        assert d.isdisjoint(a)
        assert not a.isdisjoint(b)

    @pytest.mark.parametrize('left, right', itertools.chain(
        # Do inter-levels pairs of groups fail as expected?
        itertools.permutations(component_groups, 2),
        # Do inter-levels pairs of components
        itertools.permutations(components, 2),
        # Do inter-levels pairs of components/groups fail as expected?
        _yield_mix(component_groups, components),
        # Does the function fail with inputs that are not components or groups
        ((u.atoms, 'invalid'), ),
    ))
    def test_failing_pairs(self, left, right):
        def dummy(self, other):
            return True

        with pytest.raises(TypeError):
            mda.core.groups._only_same_level(dummy)(left, right)

    @pytest.mark.parametrize('left, right', itertools.chain(
        # Groups
        _yield_sliced_groups(u, slice(0, 2), slice(1, 3)),
        # Components
        _yield_sliced_groups(u, 0, 1),
        # Mixed
        _yield_sliced_groups(u, slice(0, 2), 1),
        _yield_sliced_groups(u, 1, slice(0, 2)),
    ))
    def test_succeeding_pairs(self, left, right):
        def dummy(self, other):
            return True

        assert mda.core.groups._only_same_level(dummy)(left, right)

    def test_only_same_level_different_universes(self):
        def dummy(self, other):
            return True

        u = make_Universe()
        u2 = make_Universe()
        _only_same_level = mda.core.groups._only_same_level
        with pytest.raises(ValueError):
            _only_same_level(dummy)(u.atoms, u2.atoms)

    @pytest.mark.parametrize('op, method', ((operator.add, 'concatenate'),
                                            (operator.sub, 'difference'),
                                            (operator.and_, 'intersection'),
                                            (operator.or_, 'union'),
                                            (operator.xor, 'symmetric_difference')))
    def test_shortcut_overriding(self, op, method, level):
        def check_operator(op, method, level):
            left = getattr(u, level)[1:3]
            right = getattr(u, level)[2:4]
            assert_equal(op(left, right), getattr(left, method)(right))

        n_segments = 5
        n_residues = n_segments * 3
        n_atoms = n_residues * 3
        u = make_Universe(size=(n_atoms, n_residues, n_segments))

        check_operator(op, method, level)


class TestGroupHash(object):
    """
    Groups should be hashable.

    See issue #1397
    """
    levels = ('atoms', 'residues', 'segments')

    @pytest.fixture(params=levels)
    def level(self, request):
        return request.param

    @pytest.fixture(scope='class')
    def u(self):
        return make_Universe(size=(3, 3, 3))

    def test_hash_exists(self, u, level):
        group = getattr(u, level)
        assert isinstance(hash(group), int)

    def test_hash_equality(self, u, level):
        a = getattr(u, level)[0:-1]
        b = getattr(u, level)[0:-1]
        assert hash(a) == hash(b)

    def test_hash_difference(self, u, level):
        a = getattr(u, level)[:-1]
        b = getattr(u, level)[1:]
        assert hash(a) != hash(b)

    @pytest.mark.parametrize('level_a, level_b',
                             itertools.permutations(levels, 2))
    def test_hash_difference_cross(self, u, level_a, level_b):
        a = getattr(u, level_a)[0:-1]
        b = getattr(u, level_b)[0:-1]
        assert hash(a) != hash(b)

    def test_hash_diff_cross_universe(self, level, u):
        u2 = make_Universe(size=(3, 3, 3))
        a = getattr(u, level)
        b = getattr(u2, level)
        assert hash(a) != hash(b)


class TestAtomGroup(object):

    def test_PDB_atom_repr(self):
        u = make_Universe(extras=('altLocs', 'names', 'types', 'resnames', 'resids', 'segids'))
        assert_equal("<Atom 1: AAA of type TypeA of resname RsA, resid 1 and segid SegA and altLoc A>", u.atoms[0].__repr__())


@pytest.fixture()
def attr_universe():
    return make_Universe(('names', 'resids', 'segids'))

class TestAttributeSetting(object):
    @pytest.mark.parametrize('groupname', ['atoms', 'residues', 'segments'])
    def test_setting_group_fail(self, attr_universe, groupname):
        group = getattr(attr_universe, groupname)

        with pytest.raises(AttributeError):
            group.this = 'that'

    @pytest.mark.parametrize('groupname', ['atoms', 'residues', 'segments'])
    def test_setting_component_fails(self, attr_universe, groupname):
        component = getattr(attr_universe, groupname)[0]

        with pytest.raises(AttributeError):
            component.this = 'that'

    @pytest.mark.parametrize('attr', ['name', 'resid', 'segid'])
    @pytest.mark.parametrize('groupname', ['atoms', 'residues', 'segments'])
    def test_group_set_singular(self, attr_universe, attr, groupname):
        # this should fail as you can't set the 'name' of a 'ResidueGroup'
        group = getattr(attr_universe, groupname)
        with pytest.raises(AttributeError):
            setattr(group, attr, 24)

    def test_atom_set_name(self, attr_universe):
        attr_universe.atoms[0].name = 'this'
        assert attr_universe.atoms[0].name == 'this'

    def test_atom_set_resid(self, attr_universe):
        with pytest.raises(NotImplementedError):
            attr_universe.atoms[0].resid = 24

    def test_atom_set_segid(self, attr_universe):
        with pytest.raises(NotImplementedError):
            attr_universe.atoms[0].segid = 'this'

    def test_residue_set_name(self, attr_universe):
        with pytest.raises(AttributeError):
            attr_universe.residues[0].name = 'this'

    def test_residue_set_resid(self, attr_universe):
        attr_universe.residues[0].resid = 24
        assert attr_universe.residues[0].resid == 24

    def test_residue_set_segid(self, attr_universe):
        with pytest.raises(NotImplementedError):
            attr_universe.residues[0].segid = 'this'

    def test_segment_set_name(self, attr_universe):
        with pytest.raises(AttributeError):
            attr_universe.segments[0].name = 'this'

    def test_segment_set_resid(self, attr_universe):
        with pytest.raises(AttributeError):
            attr_universe.segments[0].resid = 24

    def test_segment_set_segid(self, attr_universe):
        attr_universe.segments[0].segid = 'this'
        assert attr_universe.segments[0].segid == 'this'

    @pytest.mark.parametrize('attr', ['names', 'resids', 'segids'])
    @pytest.mark.parametrize('groupname', ['atoms', 'residues', 'segments'])
    def test_component_set_plural(self, attr, groupname):
        # this should fail as you can't set the 'Names' of an 'Atom'
        u = make_Universe(('names', 'resids', 'segids'))
        group = getattr(u, groupname)
        comp = group[0]
        with pytest.raises(AttributeError):
            setattr(comp, attr, 24)

class TestAttributeGetting(object):
    
    @staticmethod
    @pytest.fixture()
    def universe():
        return make_Universe(extras=('masses', 'altLocs'))

    @pytest.mark.parametrize('attr', ['masses', 'altLocs'])
    def test_get_present_topattr_group(self, universe, attr):
        values = getattr(universe.atoms, attr)
        assert values is not None
    
    @pytest.mark.parametrize('attr', ['mass', 'altLoc'])
    def test_get_present_topattr_component(self, universe, attr):
        value = getattr(universe.atoms[0], attr)
        assert value is not None

    @pytest.mark.parametrize('attr,singular', [
        ('masses', 'mass'), 
        ('altLocs', 'altLoc')])
    def test_get_plural_topattr_from_component(self, universe, attr, singular):
        with pytest.raises(AttributeError) as exc:
            getattr(universe.atoms[0], attr)
        assert ('Do you mean ' + singular) in str(exc.value)
    
    @pytest.mark.parametrize('attr,singular', [
        ('masses', 'mass'), 
        ('altLocs', 'altLoc')])
    def test_get_sing_topattr_from_group(self, universe, attr, singular):
        with pytest.raises(AttributeError) as exc:
            getattr(universe.atoms, singular)
        assert ('Do you mean '+attr) in str(exc.value)
    
    @pytest.mark.parametrize('attr,singular', [
        ('elements', 'element'), 
        ('tempfactors', 'tempfactor'),
        ('bonds', 'bonds')])
    def test_get_absent_topattr_group(self, universe, attr, singular):
        with pytest.raises(NoDataError) as exc:
            getattr(universe.atoms, attr)
        assert 'does not contain '+singular in str(exc.value)

    def test_get_non_topattr(self, universe):
        with pytest.raises(AttributeError) as exc:
            universe.atoms.jabberwocky
        assert 'has no attribute' in str(exc.value)

    def test_unwrap_without_bonds(self, universe):
        with pytest.raises(NoDataError) as exc:
            universe.atoms.unwrap()
        err = ('AtomGroup.unwrap() not available; '
               'this requires Bonds')
        assert str(exc.value) == err

    def test_get_absent_attr_method(self, universe):
        with pytest.raises(NoDataError) as exc:
            universe.atoms.total_charge()
        err = ('AtomGroup.total_charge() not available; '
               'this requires charges')
        assert str(exc.value) == err

    def test_get_absent_attrprop(self, universe):
        with pytest.raises(NoDataError) as exc:
            universe.atoms.fragindices
        err = ('AtomGroup.fragindices not available; '
               'this requires bonds')
        assert str(exc.value) == err

    def test_attrprop_wrong_group(self, universe):
        with pytest.raises(AttributeError) as exc:
            universe.atoms[0].fragindices
        err = ('fragindices is a property of AtomGroup, not Atom')
        assert str(exc.value) == err

    def test_attrmethod_wrong_group(self, universe):
        with pytest.raises(AttributeError) as exc:
            universe.atoms[0].center_of_mass()
        err = ('center_of_mass() is a method of AtomGroup, not Atom')
        assert str(exc.value) == err
    
    @pytest.mark.parametrize('attr', ['altlocs', 'alt_Locs'])
    def test_wrong_name(self, universe, attr):
        with pytest.raises(AttributeError) as exc:
            getattr(universe.atoms, attr)
        err = ('AtomGroup has no attribute {}. '
               'Did you mean altLocs?').format(attr)
        assert str(exc.value) == err


class TestInitGroup(object):
    @staticmethod
    @pytest.fixture(
        params=['atoms', 'residues', 'segments']
    )
    def components(request):
        # return list of Component and container class for all three levels
        u = make_Universe()

        group = getattr(u, request.param)

        cls = {
            'atoms': mda.AtomGroup,
            'residues': mda.ResidueGroup,
            'segments': mda.SegmentGroup,
        }[request.param]

        yield (u, [group[0], group[2], group[4]], cls)

    def test_object_init(self, components):
        u, objects, cls = components

        group = cls(objects)

        assert len(group) == len(objects)
        for obj in objects:
            assert obj in group

    def test_number_init(self, components):
        u, objects, cls = components

        group = cls([0, 2, 4], u)
        for obj in objects:
            assert obj in group

    def test_VE_no_uni(self, components):
        u, objects, cls = components

        with pytest.raises(TypeError):
            cls([0, 2, 4])  # missing Universe

    def test_VE_no_uni_2(self, components):
        u, objects, cls = components

        with pytest.raises(TypeError):
            cls(0, 2, 4)  # missing Universe


class TestDecorator(object):
    @groups.check_pbc_and_unwrap
    def dummy_funtion(cls, compound="group", pbc=True, unwrap=True):
        return 0

    @pytest.mark.parametrize('compound', ('fragments', 'molecules', 'residues',
                                          'group', 'segments'))
    @pytest.mark.parametrize('pbc', (True, False))
    @pytest.mark.parametrize('unwrap', (True, False))
    def test_decorator(self, compound, pbc, unwrap):

        if compound == 'group' and pbc and unwrap:
            with pytest.raises(ValueError):
                self.dummy_funtion(compound=compound, pbc=pbc, unwrap=unwrap)
        else:
            assert_equal(self.dummy_funtion(compound=compound, pbc=pbc, unwrap=unwrap), 0)


@pytest.fixture()
def tpr():
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore",
                                message="No coordinate reader found")
        return mda.Universe(TPR)

class TestGetConnectionsAtoms(object):
    """Test Atom and AtomGroup.get_connections"""

    @pytest.mark.parametrize("typename",
                             ["bonds", "angles", "dihedrals", "impropers"])
    def test_connection_from_atom_not_outside(self, tpr, typename):
        cxns = tpr.atoms[1].get_connections(typename, outside=False)
        assert len(cxns) == 0

    @pytest.mark.parametrize("typename, n_atoms", [
        ("bonds", 1),
        ("angles", 3),
        ("dihedrals", 4),
    ])
    def test_connection_from_atom_outside(self, tpr, typename, n_atoms):
        cxns = tpr.atoms[10].get_connections(typename, outside=True)
        assert len(cxns) == n_atoms

    @pytest.mark.parametrize("typename, n_atoms", [
        ("bonds", 9),
        ("angles", 15),
        ("dihedrals", 12),
    ])
    def test_connection_from_atoms_not_outside(self, tpr, typename,
                                               n_atoms):
        ag = tpr.atoms[:10]
        cxns = ag.get_connections(typename, outside=False)
        assert len(cxns) == n_atoms
        indices = np.ravel(cxns.to_indices())
        assert np.all(np.in1d(indices, ag.indices))

    @pytest.mark.parametrize("typename, n_atoms", [
        ("bonds", 13),
        ("angles", 27),
        ("dihedrals", 38),
    ])
    def test_connection_from_atoms_outside(self, tpr, typename, n_atoms):
        ag = tpr.atoms[:10]
        cxns = ag.get_connections(typename, outside=True)
        assert len(cxns) == n_atoms
        indices = np.ravel(cxns.to_indices())
        assert not np.all(np.in1d(indices, ag.indices))

    def test_invalid_connection_error(self, tpr):
        with pytest.raises(AttributeError, match="does not contain"):
            ag = tpr.atoms[:10]
            ag.get_connections("ureybradleys")

    @pytest.mark.parametrize("outside", [True, False])
    def test_get_empty_group(self, tpr, outside):
        imp = tpr.impropers
        ag = tpr.atoms[:10]
        cxns = ag.get_connections("impropers", outside=outside)
        assert len(imp) == 0
        assert len(cxns) == 0


class TestGetConnectionsResidues(object):
    """Test Residue and ResidueGroup.get_connections"""

    @pytest.mark.parametrize("typename, n_atoms", [
        ("bonds", 9),
        ("angles", 14),
        ("dihedrals", 9),
        ("impropers", 0),
    ])
    def test_connection_from_res_not_outside(self, tpr, typename, n_atoms):
        cxns = tpr.residues[10].get_connections(typename, outside=False)
        assert len(cxns) == n_atoms

    @pytest.mark.parametrize("typename, n_atoms", [
        ("bonds", 11),
        ("angles", 22),
        ("dihedrals", 27),
        ("impropers", 0),
    ])
    def test_connection_from_res_outside(self, tpr, typename, n_atoms):
        cxns = tpr.residues[10].get_connections(typename, outside=True)
        assert len(cxns) == n_atoms

    @pytest.mark.parametrize("typename, n_atoms", [
        ("bonds", 157),
        ("angles", 290),
        ("dihedrals", 351),
    ])
    def test_connection_from_residues_not_outside(self, tpr, typename,
                                                  n_atoms):
        ag = tpr.residues[:10]
        cxns = ag.get_connections(typename, outside=False)
        assert len(cxns) == n_atoms
        indices = np.ravel(cxns.to_indices())
        assert np.all(np.in1d(indices, ag.atoms.indices))

    @pytest.mark.parametrize("typename, n_atoms", [
        ("bonds", 158),
        ("angles", 294),
        ("dihedrals", 360),
    ])
    def test_connection_from_residues_outside(self, tpr, typename, n_atoms):
        ag = tpr.residues[:10]
        cxns = ag.get_connections(typename, outside=True)
        assert len(cxns) == n_atoms
        indices = np.ravel(cxns.to_indices())
        assert not np.all(np.in1d(indices, ag.atoms.indices))

    def test_invalid_connection_error(self, tpr):
        with pytest.raises(AttributeError, match="does not contain"):
            ag = tpr.residues[:10]
            ag.get_connections("ureybradleys")

    @pytest.mark.parametrize("outside", [True, False])
    def test_get_empty_group(self, tpr, outside):
        imp = tpr.impropers
        ag = tpr.residues[:10]
        cxns = ag.get_connections("impropers", outside=outside)
        assert len(imp) == 0
        assert len(cxns) == 0


@pytest.mark.parametrize("typename, n_inside", [
    ("intra_bonds", 9),
    ("intra_angles", 15),
    ("intra_dihedrals", 12),
])
def test_topologygroup_gets_connections_inside(tpr, typename, n_inside):
    ag = tpr.atoms[:10]
    cxns = getattr(ag, typename)
    assert len(cxns) == n_inside
    indices = np.ravel(cxns.to_indices())
    assert np.all(np.in1d(indices, ag.indices))


@pytest.mark.parametrize("typename, n_outside", [
    ("bonds", 13),
    ("angles", 27),
    ("dihedrals", 38),
])
def test_topologygroup_gets_connections_outside(tpr, typename, n_outside):
    ag = tpr.atoms[:10]
    cxns = getattr(ag, typename)
    assert len(cxns) == n_outside



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest
from numpy.testing import assert_equal
import functools

import numpy as np

from MDAnalysis.tests.datafiles import (
    TPR,
    TPR400, TPR402, TPR403, TPR404, TPR405, TPR406, TPR407,
    TPR450, TPR451, TPR452, TPR453, TPR454, TPR455, TPR455Double,
    TPR460, TPR461, TPR502, TPR504, TPR505, TPR510, TPR510_bonded,
    TPR2016, TPR2018, TPR2019B3, TPR2020B2, TPR2020, TPR2020Double,
    TPR2021, TPR2021Double,
    TPR2016_bonded, TPR2018_bonded, TPR2019B3_bonded,
    TPR2020B2_bonded, TPR2020_bonded, TPR2020_double_bonded,
    TPR2021_bonded, TPR2021_double_bonded, TPR334_bonded,
    TPR_EXTRA_2021, TPR_EXTRA_2020, TPR_EXTRA_2018,
    TPR_EXTRA_2016, TPR_EXTRA_407,
    XTC,
)
from MDAnalysisTests.topology.base import ParserBase
import MDAnalysis.topology.TPRParser
import MDAnalysis as mda

BONDED_TPRS = (
    TPR510_bonded,
    TPR2016_bonded,
    TPR2018_bonded,
    TPR2019B3_bonded,
    TPR2021_bonded,
    TPR2021_double_bonded,
    TPR2020_bonded,
    TPR2020_double_bonded,
    TPR_EXTRA_2021,
    TPR_EXTRA_2020,
    TPR_EXTRA_2018,
    TPR_EXTRA_2016,
    TPR_EXTRA_407,
)


class TPRAttrs(ParserBase):
    parser = MDAnalysis.topology.TPRParser.TPRParser
    expected_attrs = ['ids', 'names', 'elements',
                      'resids', 'resnames',
                      'moltypes', 'molnums', 'charges',
                      'bonds', 'angles', 'dihedrals', 'impropers']

    def test_moltypes(self, top):
        moltypes = top.moltypes.values
        assert_equal(moltypes, self.ref_moltypes)

    def test_molnums(self, top):
        molnums = top.molnums.values
        assert_equal(molnums, self.ref_molnums)
        assert molnums.dtype == np.intp


class TestTPR(TPRAttrs):
    """
    this test the data/adk_oplsaa.tpr which is of tpx version 58
    """
    expected_n_atoms = 47681
    expected_n_residues = 11302
    expected_n_segments = 3
    ref_moltypes = np.array(['AKeco'] * 214 + ['SOL'] * 11084 + ['NA+'] * 4,
                            dtype=object)
    ref_molnums = np.array([0] * 214 + list(range(1, 1 + 11084 + 4)))

    @pytest.fixture()
    def filename(self):
        return TPR


# The follow test the same system grompped by different version of gromacs
# FORMAT: TPRABC, where numbers ABC indicates the version of gromacs that
# generates the corresponding tpr file
class TestTPRGromacsVersions(TPRAttrs):
    expected_n_atoms = 2263
    expected_n_residues = 230
    expected_n_segments = 2
    ref_moltypes = np.array(['Protein_A'] * 129 + ['SOL'] * 101, dtype=object)
    ref_molnums = np.array([0] * 129 + list(range(1, 1 + 101)))

    @pytest.fixture(params=[TPR400, TPR402, TPR403, TPR404, TPR405, TPR406,
                            TPR407, TPR450, TPR451, TPR452, TPR453, TPR454,
                            TPR455, TPR502, TPR504, TPR505, TPR510, TPR2016,
                            TPR2018, TPR2019B3, TPR2020, TPR2020Double,
                            TPR2021, TPR2021Double])
    def filename(self, request):
        return request.param


class TestTPRDouble(TPRAttrs):
    expected_n_atoms = 21692
    expected_n_residues = 4352
    expected_n_segments = 7
    ref_moltypes = np.array(['DOPC'] * 21 + ['DPPC'] * 10 + ['CHOL'] * 3
                            + ['DOPC'] * 21 + ['DPPC'] * 10 + ['CHOL'] * 3
                            + ['SOL'] * 4284,
                            dtype=object)
    ref_molnums = np.arange(4352)

    @pytest.fixture()
    def filename(self):
        return TPR455Double


class TestTPR46x(TPRAttrs):
    expected_n_atoms = 44052
    expected_n_residues = 10712
    expected_n_segments = 8
    ref_moltypes = np.array(['Protein_A'] * 27 + ['Protein_B'] * 27
                            + ['Protein_C'] * 27 + ['Protein_D'] * 27
                            + ['Protein_E'] * 27
                            + ['SOL'] * 10530 + ['NA+'] * 26 + ['CL-'] * 21,
                            dtype=object)
    ref_molnums = np.array([0] * 27 + [1] * 27 + [2] * 27 + [3] * 27 + [4] * 27
                           + list(range(5, 5 + 10530 + 26 + 21)))

    @pytest.fixture(params=[TPR460, TPR461])
    def filename(self, request):
        return request.param


def _test_is_in_topology(name, elements, topology_path, topology_section):
    """
    Test if an interaction appears as expected in the topology
    """
    post_40_potentials = {
        'RESTRAINTPOT', 'RESTRANGLES', 'RESTRDIHS', 'CBTDIHS', 'PIDIHS',
    }
    if name in post_40_potentials and topology_path == TPR_EXTRA_407:
        # The potential is not yet implemented in this version of gromacs
        return
    parser = MDAnalysis.topology.TPRParser.TPRParser(topology_path)
    top = parser.parse()
    for element in elements:
        assert element in getattr(top, topology_section).values, \
            'Interaction type "{}" not found'.format(name)


@pytest.mark.parametrize('topology', BONDED_TPRS)
@pytest.mark.parametrize('bond', (
        ('BONDS', [(0, 1)]),
        ('G96BONDS', [(1, 2)]),
        ('MORSE', [(2, 3)]),
        ('CUBICBONDS', [(3, 4)]),
        ('CONNBONDS', [(4, 5)]),
        ('HARMONIC', [(5, 6)]),
        ('FENEBONDS', [(6, 7)]),
        ('RESTRAINTPOT', [(7, 8)]),
        ('TABBONDS', [(8, 9)]),
        ('TABBONDSNC', [(9, 10)]),
        ('CONSTR', [(10, 11)]),
        ('CONSTRNC', [(11, 12)]),
))
def test_all_bonds(topology, bond):
    """Test that all bond types are parsed as expected"""
    bond_type_in_topology = functools.partial(_test_is_in_topology,
                                              topology_section='bonds')
    bond_type, elements = bond
    bond_type_in_topology(bond_type, elements, topology)


@pytest.mark.parametrize('topology', BONDED_TPRS)
@pytest.mark.parametrize('angle', (
    ('ANGLES', [(0, 1, 2)]),
    ('G96ANGLES', [(1, 2, 3)]),
    ('CROSS_BOND_BOND', [(2, 3, 4)]),
    ('CROSS_BOND_ANGLE', [(3, 4, 5)]),
    ('UREY_BRADLEY', [(4, 5, 6)]),
    ('QANGLES', [(5, 6, 7)]),
    ('RESTRANGLES', [(6, 7, 8)]),
    ('TABANGLES', [(7, 8, 9)]),
))
def test_all_angles(topology, angle):
    angle_type_in_topology = functools.partial(_test_is_in_topology,
                                               topology_section='angles')
    angle_type, elements = angle
    angle_type_in_topology(angle_type, elements, topology)


@pytest.mark.parametrize('topology', BONDED_TPRS)
@pytest.mark.parametrize('dih', (
        ('PDIHS', [(0, 1, 2, 3), (1, 2, 3, 4), (7, 8, 9, 10)]),
        ('RBDIHS', [(4, 5, 6, 7)]),
        ('RESTRDIHS', [(8, 9, 10, 11)]),
        ('CBTDIHS', [(9, 10, 11, 12)]),
        ('FOURDIHS', [(6, 7, 8, 9)]),
        ('TABDIHS', [(10, 11, 12, 13)]),
))
def test_all_dihedrals(topology, dih):
    dih_type_in_topology = functools.partial(_test_is_in_topology,
                                             topology_section='dihedrals')
    dih_type, elements = dih
    dih_type_in_topology(dih_type, elements, topology)


@pytest.mark.parametrize('topology', BONDED_TPRS)
@pytest.mark.parametrize('impr', (
    ('IDIHS', [(2, 3, 4, 5), (3, 4, 5, 6)]),
    ('PIDIHS', [(5, 6, 7, 8)])
))
def test_all_impropers(topology, impr):
    impr_type_in_topology = functools.partial(_test_is_in_topology,
                                              topology_section='impropers')

    impr_type, elements = impr
    impr_type_in_topology(impr_type, elements, topology)


@pytest.fixture(params=(
    TPR400, TPR402, TPR403, TPR404, TPR405, TPR406, TPR407, TPR450, TPR451,
    TPR452, TPR453, TPR454, TPR502, TPR504, TPR505, TPR510, TPR2016, TPR2018,
))
def bonds_water(request):
    parser = MDAnalysis.topology.TPRParser.TPRParser(request.param).parse()
    # The index of the first water atom is 1960
    first = 1960
    bonds = [
        bond
        for bond in parser.bonds.values
        if bond[0] >= first and bond[1] >= first
    ]
    return bonds


def test_settle(bonds_water):
    # There are 101 water molecule with 2 bonds each
    assert len(bonds_water) == 202
    # The last index corresponds to the last water atom
    assert bonds_water[-1][1] == 2262


@pytest.mark.parametrize('tpr_path, expected_exception', (
    (TPR2020B2, IOError),  # Gromacs 2020 beta see issue #2428
    (TPR2020B2_bonded, IOError),  # Gromacs 2020 beta see issue #2428
    (TPR334_bonded, NotImplementedError),  # Too old
    (XTC, IOError),  # Not a TPR file
))
def test_fail_for_unsupported_files(tpr_path, expected_exception):
    parser = MDAnalysis.topology.TPRParser.TPRParser(tpr_path)
    with pytest.raises(expected_exception):
        parser.parse()


@pytest.mark.parametrize('tpr_path', BONDED_TPRS)
def test_no_elements(tpr_path):
    """
    If the TPR does not contain element information, the element topology
    attribute is not defined.
    """
    parser = MDAnalysis.topology.TPRParser.TPRParser(tpr_path)
    topology = parser.parse()
    with pytest.raises(AttributeError):
        _ = topology.elements


def test_elements():
    tpr_path = TPR
    parser = MDAnalysis.topology.TPRParser.TPRParser(tpr_path)
    topology = parser.parse()
    reference = np.array((
         'H,C,H,H,C,H,H,H,C,H,H,H,C,O,N,H,C,H,C,H,H,C,H,C,H,H,H,C,H,H,'
         'H,C,O,N,H,C,H,H,C,O,O,O,H,H,,O,H,H,,O,H,H,,O,H,H,,O,H,H,,O,H'
         ',H,,O,H,H,,O,H,H,,O,H,H,,O,H,H,,O,H,H,,O,H,H,,O,H,H,,O,H,H,,'
         'O,H,H'
    ).split(','), dtype=object)
    assert_equal(topology.elements.values[3300:3400], reference)
    reference = np.array([
        'O', 'H', 'H', '', 'O', 'H', 'H', '', 'O', 'H', 'H', '', 'O', 'H',
        'H', '', 'Na', 'Na', 'Na', 'Na',
    ], dtype=object)
    assert_equal(topology.elements.values[-20:], reference)


@pytest.mark.parametrize("resid_from_one,resid_addition", [
    (False, 0),
    (True, 1),  # status quo for 2.x
    ])
def test_resids(resid_from_one, resid_addition):
    u = mda.Universe(TPR, tpr_resid_from_one=resid_from_one)
    resids = np.arange(len(u.residues)) + resid_addition
    assert_equal(u.residues.resids, resids,
                 err_msg="tpr_resid_from_one kwarg not switching resids")



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- http://www.MDAnalysis.org
# Copyright (c) 2006-2015 Naveen Michaud-Agrawal, Elizabeth J. Denning, Oliver
# Beckstein and contributors (see AUTHORS for the full list)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pickle

from collections import namedtuple
import os
import sys
import string
import struct
import platform

import hypothesis.strategies as strategies
from hypothesis import example, given
import hypothesis
import numpy as np

from numpy.testing import (assert_array_almost_equal, assert_equal,
                           assert_array_equal, assert_almost_equal)

from MDAnalysis.lib.formats.libdcd import DCDFile, DCD_IS_CHARMM, DCD_HAS_EXTRA_BLOCK

from MDAnalysisTests.datafiles import (
    DCD, DCD_NAMD_TRICLINIC, legacy_DCD_ADK_coords, legacy_DCD_NAMD_coords,
    legacy_DCD_c36_coords, DCD_TRICLINIC)

import pytest


@pytest.mark.parametrize("dcdfile, is_periodic",
                         [(DCD, False), (DCD_NAMD_TRICLINIC, True),
                          (DCD_TRICLINIC, True)])
def test_is_periodic(dcdfile, is_periodic):
    with DCDFile(dcdfile) as f:
        assert f.is_periodic == is_periodic


@pytest.mark.parametrize("dcdfile, natoms", [(DCD, 3341), (DCD_NAMD_TRICLINIC,
                                                           5545),
                                             (DCD_TRICLINIC, 375)])
def test_read_coordsshape(dcdfile, natoms):
    # confirm shape of coordinate data against result from previous
    # MDAnalysis implementation of DCD file handling
    with DCDFile(dcdfile) as dcd:
        dcd_frame = dcd.read()
    xyz = dcd_frame[0]
    assert xyz.shape == (natoms, 3)


@pytest.mark.parametrize(
    "dcdfile, unit_cell",
    [(DCD, [0., 90., 0., 90., 90., 0.]),
     (DCD_NAMD_TRICLINIC, [38.42659378, 0.499563, 38.393102, 0., 0., 44.7598]),
     (DCD_TRICLINIC,
      [30.841836, 14.578635, 31.780088, 9.626323, -2.60815, 32.67009])])
def test_read_unit_cell(dcdfile, unit_cell):
    # confirm unit cell read against result from previous
    # MDAnalysis implementation of DCD file handling
    with DCDFile(dcdfile) as dcd:
        dcd_frame = dcd.read()
    assert_array_almost_equal(dcd_frame.unitcell, unit_cell)


def test_seek_over_max():
    with DCDFile(DCD) as dcd:
        with pytest.raises(EOFError):
            dcd.seek(102)


@pytest.fixture
def dcd():
    with DCDFile(DCD) as dcd:
        yield dcd


def _assert_compare_readers(old_reader, new_reader):
    #  same as next(old_reader)
    frame = old_reader.read()
    #  same as next(new_reader)
    new_frame = new_reader.read()

    assert old_reader.fname == new_reader.fname
    assert old_reader.tell() == new_reader.tell()
    assert_almost_equal(frame.xyz, new_frame.xyz)
    assert_almost_equal(frame.unitcell, new_frame.unitcell)


def test_pickle(dcd):
    mid = len(dcd) // 2
    dcd.seek(mid)
    new_dcd = pickle.loads(pickle.dumps(dcd))
    _assert_compare_readers(dcd, new_dcd)


def test_pickle_last(dcd):
    #  This is the file state when DCDReader is in its last frame.
    #  (Issue #2878)

    dcd.seek(len(dcd) - 1)
    _ = dcd.read()
    new_dcd = pickle.loads(pickle.dumps(dcd))

    assert dcd.fname == new_dcd.fname
    assert dcd.tell() == new_dcd.tell()
    with pytest.raises(StopIteration):
        new_dcd.read()


def test_pickle_closed(dcd):
    dcd.seek(len(dcd) - 1)
    dcd.close()
    new_dcd = pickle.loads(pickle.dumps(dcd))

    assert dcd.fname == new_dcd.fname
    assert dcd.tell() != new_dcd.tell()


def test_pickle_after_read(dcd):
    _ = dcd.read()
    new_dcd = pickle.loads(pickle.dumps(dcd))
    _assert_compare_readers(dcd, new_dcd)


def test_pickle_immediately(dcd):
    new_dcd = pickle.loads(pickle.dumps(dcd))

    assert dcd.fname == new_dcd.fname
    assert dcd.tell() == new_dcd.tell()


@pytest.mark.parametrize("new_frame", (10, 42, 21))
def test_seek_normal(new_frame, dcd):
    # frame seek within range is tested
    dcd.seek(new_frame)
    assert dcd.tell() == new_frame


def test_seek_negative(dcd):
    with pytest.raises(IOError):
        dcd.seek(-78)


def test_iteration(dcd):
    num_iters = 10
    for _ in range(num_iters):
        dcd.__next__()
    assert dcd.tell() == num_iters


def test_open_wrong_mode():
    with pytest.raises(IOError):
        DCDFile(DCD, 'e')


def test_raise_not_existing():
    with pytest.raises(IOError):
        DCDFile('foo')


def test_zero_based_frames_counting(dcd):
    assert dcd.tell() == 0


@pytest.mark.parametrize("dcdfile, natoms", [(DCD, 3341), (DCD_NAMD_TRICLINIC,
                                                           5545),
                                             (DCD_TRICLINIC, 375)])
def test_natoms(dcdfile, natoms):
    with DCDFile(dcdfile) as dcd:
        assert dcd.header['natoms'] == natoms


def test_read_closed(dcd):
    dcd.close()
    with pytest.raises(IOError):
        dcd.read()


@pytest.mark.parametrize("dcdfile, nframes", [(DCD, 98), (DCD_NAMD_TRICLINIC,
                                                          1), (DCD_TRICLINIC,
                                                               10)])
def test_length_traj(dcdfile, nframes):
    with DCDFile(dcdfile) as dcd:
        assert len(dcd) == nframes


def test_read_write_mode_file(tmpdir):
    fname = str(tmpdir.join('foo'))
    with DCDFile(fname, 'w') as f:
        with pytest.raises(IOError):
            f.read()


def test_iterating_twice(dcd):
    with dcd as f:
        for i, _ in enumerate(f):
            assert_equal(i + 1, f.tell())
        # second iteration should work from start again
        for i, _ in enumerate(f):
            assert_equal(i + 1, f.tell())


DCD_HEADER = '''* DIMS ADK SEQUENCE FOR PORE PROGRAM                                            * WRITTEN BY LIZ DENNING (6.2008)                                               *  DATE:     6/ 6/ 8     17:23:56      CREATED BY USER: denniej0                '''
DCD_NAMD_TRICLINIC_HEADER = 'Created by DCD pluginREMARKS Created 06 July, 2014 at 17:29Y5~CORD,'
DCD_TRICLINIC_HEADER = '* CHARMM TRICLINIC BOX TESTING                                                  * (OLIVER BECKSTEIN 2014)                                                       * BASED ON NPTDYN.INP : SCOTT FELLER, NIH, 7/15/95                              * TEST EXTENDED SYSTEM CONSTANT PRESSURE AND TEMPERATURE                        * DYNAMICS WITH WATER BOX.                                                      *  DATE:     7/ 7/14     13:59:46      CREATED BY USER: oliver                  '


@pytest.mark.parametrize("dcdfile, remarks",
                         ((DCD, DCD_HEADER), (DCD_NAMD_TRICLINIC,
                                              DCD_NAMD_TRICLINIC_HEADER),
                          (DCD_TRICLINIC, DCD_TRICLINIC_HEADER)))
def test_header_remarks(dcdfile, remarks):
    # confirm correct header remarks section reading
    with DCDFile(dcdfile) as f:
        assert len(f.header['remarks']) == len(remarks)


@pytest.mark.parametrize("dcdfile, legacy_data, frames",
                         ((DCD, legacy_DCD_ADK_coords, [5, 29]),
                          (DCD_NAMD_TRICLINIC, legacy_DCD_NAMD_coords, [0]),
                          (DCD_TRICLINIC, legacy_DCD_c36_coords, [1, 4])))
def test_read_coord_values(dcdfile, legacy_data, frames):
    # test the actual values of coordinates read in versus
    # stored values read in by the legacy DCD handling framework
    # to reduce repo storage burden, we only compare for a few
    # randomly selected frames
    legacy = np.load(legacy_data)
    with DCDFile(dcdfile) as dcd:
        for index, frame_num in enumerate(frames):
            dcd.seek(frame_num)
            actual_coords = dcd.read()[0]
            desired_coords = legacy[index]
            assert_array_equal(actual_coords, desired_coords)


@pytest.mark.parametrize("dcdfile, legacy_data, frame_idx",
                         ((DCD, legacy_DCD_ADK_coords, [5, 29]),
                          (DCD_NAMD_TRICLINIC, legacy_DCD_NAMD_coords, [0]),
                          (DCD_TRICLINIC, legacy_DCD_c36_coords, [1, 4])))
def test_readframes(dcdfile, legacy_data, frame_idx):
    legacy = np.load(legacy_data)
    with DCDFile(dcdfile) as dcd:
        frames = dcd.readframes()
        xyz = frames.xyz
        assert_equal(len(xyz), len(dcd))
        for index, frame_num in enumerate(frame_idx):
            assert_array_almost_equal(xyz[frame_num], legacy[index])


def test_write_header(tmpdir):
    # test that _write_header() can produce a very crude
    # header for a new / empty file
    testfile = str(tmpdir.join('test.dcd'))
    with DCDFile(testfile, 'w') as dcd:
        dcd.write_header(
            remarks='Crazy!',
            natoms=22,
            istart=12,
            nsavc=10,
            delta=0.02,
            is_periodic=1)

    with DCDFile(testfile) as dcd:
        header = dcd.header
        assert header['remarks'] == 'Crazy!'
        assert header['natoms'] == 22
        assert header['istart'] == 12
        assert header['is_periodic'] == 1
        assert header['nsavc'] == 10
        assert np.allclose(header['delta'], .02)

    # we also check the bytes written directly.
    with open(testfile, 'rb') as fh:
        header_bytes = fh.read()
    # check for magic number
    assert struct.unpack('i', header_bytes[:4])[0] == 84
    # magic number should be written again before remark section
    assert struct.unpack('i', header_bytes[88:92])[0] == 84
    # length of remark section. We hard code this to 244 right now
    assert struct.unpack('i', header_bytes[92:96])[0] == 244
    # say we have 3 block of length 80
    assert struct.unpack('i', header_bytes[96:100])[0] == 3
    # after the remark section the length should be reported again
    assert struct.unpack('i', header_bytes[340:344])[0] == 244
    # this is a magic number as far as I see
    assert struct.unpack('i', header_bytes[344:348])[0] == 4


def test_write_no_header(tmpdir):
    fname = str(tmpdir.join('test.dcd'))
    with DCDFile(fname, 'w') as dcd:
        with pytest.raises(IOError):
            dcd.write(np.ones(3), np.ones(6))


def test_write_header_twice(tmpdir):
    # an IOError should be raised if a duplicate
    # header writing is attempted

    header = {
        "remarks": 'Crazy!',
        "natoms": 22,
        "istart": 12,
        "nsavc": 10,
        "delta": 0.02,
        "is_periodic": 1
    }

    fname = str(tmpdir.join('test.dcd'))
    with DCDFile(fname, 'w') as dcd:
        dcd.write_header(**header)
        with pytest.raises(IOError):
            dcd.write_header(**header)


def test_write_header_wrong_mode(dcd):
    # an exception should be raised on any attempt to use
    # write_header with a DCDFile object in 'r' mode
    with pytest.raises(IOError):
        dcd.write_header(
            remarks='Crazy!',
            natoms=22,
            istart=12,
            nsavc=10,
            delta=0.02,
            is_periodic=1)


def test_write_mode(dcd):
    # ensure that writing of DCD files only occurs with properly
    # opened files
    with pytest.raises(IOError):
        dcd.write(xyz=np.zeros((3, 3)), box=np.zeros(6, dtype=np.float64))


def write_dcd(in_name, out_name, remarks='testing', header=None):
    with DCDFile(in_name) as f_in, DCDFile(out_name, 'w') as f_out:
        if header is None:
            header = f_in.header
        f_out.write_header(**header)
        for frame in f_in:
            f_out.write(xyz=frame.xyz, box=frame.unitcell)


@pytest.mark.xfail((os.name == 'nt'
                    and sys.maxsize <= 2**32) or
                    platform.machine() == 'aarch64',
                   reason="occasional fail on 32-bit windows and ARM")
# occasionally fails due to unreliable test timings
@hypothesis.settings(deadline=None)  # see Issue 3096
@given(remarks=strategies.text(
    alphabet=string.printable, min_size=0,
    max_size=239))  # handle the printable ASCII strings
@example(remarks='')
def test_written_remarks_property(remarks, tmpdir_factory):
    # property based testing for writing of a wide range of string
    # values to REMARKS field
    dcd = DCDFile(DCD)
    dirname = str(id(remarks)) + "_"
    testfile = str(tmpdir_factory.mktemp(dirname).join('test.dcd'))
    header = dcd.header
    header['remarks'] = remarks
    write_dcd(DCD, testfile, header=header)
    expected_remarks = remarks
    with DCDFile(testfile) as f:
        assert f.header['remarks'] == expected_remarks


@pytest.fixture(scope='session')
def written_dcd(tmpdir_factory):
    with DCDFile(DCD) as dcd:
        header = dcd.header
    testfile = tmpdir_factory.mktemp('dcd').join('test.dcd')
    testfile = str(testfile)
    write_dcd(DCD, testfile)
    Result = namedtuple("Result", "testfile, header, orgfile")
    # throw away last char we didn't save due to null termination
    header['remarks'] = header['remarks'][:-1]
    return Result(testfile, header, DCD)


def test_written_header(written_dcd):
    header = written_dcd.header
    with DCDFile(written_dcd.testfile) as dcd:
        dcdheader = dcd.header
        assert dcdheader == header


def test_written_num_frames(written_dcd):
    with DCDFile(written_dcd.testfile) as dcd, DCDFile(
            written_dcd.orgfile) as other:
        assert len(dcd) == len(other)


def test_written_dcd_coordinate_data_shape(written_dcd):
    with DCDFile(written_dcd.testfile) as dcd, DCDFile(
            written_dcd.orgfile) as other:
        for frame, other_frame in zip(dcd, other):
            assert frame.xyz.shape == other_frame.xyz.shape


def test_written_seek(written_dcd):
    # ensure that we can seek properly on written DCD file
    with DCDFile(written_dcd.testfile) as f:
        f.seek(40)
        assert_equal(f.tell(), 40)


def test_written_coord_match(written_dcd):
    with DCDFile(written_dcd.testfile) as test, DCDFile(
            written_dcd.orgfile) as ref:
        for frame, o_frame in zip(test, ref):
            assert_array_almost_equal(frame.xyz, o_frame.xyz)


def test_written_unit_cell(written_dcd):
    with DCDFile(written_dcd.testfile) as test, DCDFile(
            written_dcd.orgfile) as ref:
        for frame, o_frame in zip(test, ref):
            assert_array_almost_equal(frame.unitcell, o_frame.unitcell)


@pytest.mark.parametrize("dtype", (np.int32, np.int64, np.float32, np.float64,
                                   int, float))
def test_write_all_dtypes(tmpdir, dtype):
    fname = str(tmpdir.join('foo.dcd'))
    with DCDFile(fname, 'w') as out:
        natoms = 10
        xyz = np.ones((natoms, 3), dtype=dtype)
        box = np.ones(6, dtype=dtype)
        out.write_header(
            remarks='test',
            natoms=natoms,
            is_periodic=1,
            delta=1,
            nsavc=1,
            istart=1)
        out.write(xyz=xyz, box=box)


@pytest.mark.parametrize("array_like", (np.array, list))
def test_write_array_like(tmpdir, array_like):
    fname = str(tmpdir.join('foo.dcd'))
    with DCDFile(fname, 'w') as out:
        natoms = 10
        xyz = array_like([[1, 1, 1] for i in range(natoms)])
        box = array_like([i for i in range(6)])
        out.write_header(
            remarks='test',
            natoms=natoms,
            is_periodic=1,
            delta=1,
            nsavc=1,
            istart=1)
        out.write(xyz=xyz, box=box)


def test_write_wrong_shape_xyz(tmpdir):
    fname = str(tmpdir.join('foo.dcd'))
    with DCDFile(fname, 'w') as out:
        natoms = 10
        xyz = np.ones((natoms + 1, 3))
        box = np.ones(6)
        out.write_header(
            remarks='test',
            natoms=natoms,
            is_periodic=1,
            delta=1,
            nsavc=1,
            istart=1)
        with pytest.raises(ValueError):
            out.write(xyz=xyz, box=box)


def test_write_wrong_shape_box(tmpdir):
    fname = str(tmpdir.join('foo.dcd'))
    with DCDFile(fname, 'w') as out:
        natoms = 10
        xyz = np.ones((natoms, 3))
        box = np.ones(7)
        out.write_header(
            remarks='test',
            natoms=natoms,
            is_periodic=1,
            delta=1,
            nsavc=1,
            istart=1)
        with pytest.raises(ValueError):
            out.write(xyz=xyz, box=box)


@pytest.mark.parametrize("dcdfile", (DCD, DCD_TRICLINIC, DCD_NAMD_TRICLINIC))
def test_relative_frame_sizes(dcdfile):
    # the first frame of a DCD file should always be >= in size
    # to subsequent frames, as the first frame contains the same
    # atoms + (optional) fixed atoms
    with DCDFile(dcdfile) as dcd:
        first_frame_size = dcd._firstframesize
        general_frame_size = dcd._framesize

    assert first_frame_size >= general_frame_size


@pytest.mark.parametrize("dcdfile", (DCD, DCD_TRICLINIC, DCD_NAMD_TRICLINIC))
def test_file_size_breakdown(dcdfile):
    # the size of a DCD file is equivalent to the sum of the header
    # size, first frame size, and (N - 1 frames) * size per general
    # frame

    expected = os.path.getsize(dcdfile)
    with DCDFile(dcdfile) as dcd:
        actual = dcd._header_size + dcd._firstframesize + (
            (dcd.n_frames - 1) * dcd._framesize)
    assert actual == expected


@pytest.mark.parametrize("dcdfile", (DCD, DCD_TRICLINIC, DCD_NAMD_TRICLINIC))
def test_nframessize_int(dcdfile):
    # require that the (nframessize / framesize) value used by DCDFile
    # is an integer (because nframessize / framesize + 1 = total frames,
    # which must also be an int)
    filesize = os.path.getsize(dcdfile)
    with DCDFile(dcdfile) as dcd:
        nframessize = filesize - dcd._header_size - dcd._firstframesize
        assert float(nframessize) % float(dcd._framesize) == 0


@pytest.mark.parametrize(
    "slice, length",
    [([None, None, None], 98), ([0, None, None], 98), ([None, 98, None], 98),
     ([None, None, 1], 98), ([None, None, -1], 98), ([2, 6, 2], 2),
     ([0, 10, None], 10), ([2, 10, None], 8), ([0, 1, 1], 1), ([1, 1, 1], 0),
     ([1, 2, 1], 1), ([1, 2, 2], 1), ([1, 4, 2], 2), ([1, 4, 4], 1), ([
         0, 5, 5
     ], 1), ([3, 5, 1], 2), ([4, 0, -1], 4), ([5, 0, -2], 3), ([5, 0, -4], 2)])
def test_readframes_slices(slice, length, dcd):
    start, stop, step = slice
    allframes = dcd.readframes().xyz
    frames = dcd.readframes(start=start, stop=stop, step=step)
    xyz = frames.xyz
    assert len(xyz) == length
    assert_array_almost_equal(xyz, allframes[start:stop:step])


@pytest.mark.parametrize("order, shape", (
    ('fac', (98, 3341, 3)),
    ('fca', (98, 3, 3341)),
    ('afc', (3341, 98, 3)),
    ('acf', (3341, 3, 98)),
    ('caf', (3, 3341, 98)),
    ('cfa', (3, 98, 3341)), ))
def test_readframes_order(order, shape, dcd):
    x = dcd.readframes(order=order).xyz
    assert x.shape == shape


@pytest.mark.parametrize("indices", [[1, 2, 3, 4], [5, 10, 15, 19],
                                     [9, 4, 2, 0, 50]])
def test_readframes_atomindices(indices, dcd):
    allframes = dcd.readframes(order='afc').xyz
    frames = dcd.readframes(indices=indices, order='afc')
    xyz = frames.xyz
    assert len(xyz) == len(indices)
    assert_array_almost_equal(xyz, allframes[indices])


def test_write_random_unitcell(tmpdir):
    testname = str(tmpdir.join('test.dcd'))
    rstate = np.random.RandomState(1178083)
    random_unitcells = rstate.uniform(high=80, size=(98, 6)).astype(np.float64)

    with DCDFile(DCD) as f_in, DCDFile(testname, 'w') as f_out:
        header = f_in.header
        header['is_periodic'] = True
        f_out.write_header(**header)
        for index, frame in enumerate(f_in):
            f_out.write(xyz=frame.xyz, box=random_unitcells[index])

    with DCDFile(testname) as test:
        for index, frame in enumerate(test):
            assert_array_almost_equal(frame.unitcell, random_unitcells[index])



"""Tests for core.groups.requires decorator

"""
import numpy as np

import pytest
from MDAnalysis.core.groups import requires
from MDAnalysis import NoDataError

from MDAnalysisTests import make_Universe


class TestRequires(object):
    def test_requires_failure_singular(self):
        @requires('masses')
        def mass_multiplier(ag1, ag2, scalar):
            return (ag1.masses + ag2.masses) * scalar

        u = make_Universe(('charges',))
        with pytest.raises(NoDataError):
            mass_multiplier(u.atoms[:10], u.atoms[20:30], 4.0)


    def test_requires_failure_multiple(self):
        @requires('masses', 'charges')
        def mass_multiplier(ag1, ag2, scalar):
            return (ag1.masses + ag2.charges) * scalar
        

        u = make_Universe(('masses', 'types'))
        with pytest.raises(NoDataError):
            mass_multiplier(u.atoms[:10], u.atoms[20:30], 4.0)

    def test_requires_success(self):
        @requires('masses')
        def mass_multiplier(ag1, ag2, scalar):
            return (ag1.masses + ag2.masses) * scalar

        u = make_Universe(('masses',))

        result = mass_multiplier(u.atoms[:10], u.atoms[20:30], 4.0)

        assert isinstance(result, np.ndarray)

    def test_failure_errormessage(self):
        # failures should list all required attributes not
        # just the first one
        @requires('cats', 'dogs', 'frogs')
        def animal_print(ag):
            return len(ag.cats), len(ag.dogs), len(ag.frogs)

        u = make_Universe()
        try:
            animal_print(u.atoms)
        except NoDataError as e:
            message = e.args[0]
            # Test function name gets returned (for debug)
            assert 'animal_print' in message
            assert 'cats' in message
            assert 'dogs' in message
            assert 'frogs' in message
        else:
            pytest.fail(msg="Should raise NoDataError")



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
from numpy.testing import assert_equal

import pytest
import bz2
import MDAnalysis as mda

from MDAnalysisTests.topology.base import ParserBase
from MDAnalysisTests.datafiles import (
    PSF,
    PSF_nosegid,
    PSF_notop,
    PSF_NAMD,
    XYZ_psf,
    XYZ,
)

class PSFBase(ParserBase):
    parser = mda.topology.PSFParser.PSFParser
    expected_attrs = ['ids', 'names', 'types', 'masses',
                      'charges',
                      'resids', 'resnames',
                      'segids',
                      'bonds', 'angles', 'dihedrals', 'impropers']


class TestPSFParser(PSFBase):
    """
    Based on small PDB with AdK (:data:`PDB_small`).
    """
    ref_filename = PSF
    expected_n_atoms = 3341
    expected_n_residues = 214
    expected_n_segments = 1

    @pytest.fixture(params=['uncompressed', 'bz2'])
    def filename(self, request, tmpdir):
        if request.param == 'uncompressed':
            return self.ref_filename
        else:
            fn = str(tmpdir.join('file.psf.bz2'))
            with open(self.ref_filename, 'rb') as f:
                stuff = f.read()
            buf = bz2.compress(stuff)
            with open(fn, 'wb') as out:
                out.write(buf)
            return fn

    def test_bonds_total_counts(self, top):
        assert len(top.bonds.values) == 3365

    def test_bonds_atom_counts(self, filename):
        u = mda.Universe(filename)
        assert len(u.atoms[[0]].bonds) == 4
        assert len(u.atoms[[42]].bonds) == 1

    def test_bonds_identity(self, top):
        vals = top.bonds.values
        for b in ((0, 1), (0, 2), (0, 3), (0, 4)):
            assert (b in vals) or (b[::-1] in vals)

    def test_angles_total_counts(self, top):
        assert len(top.angles.values) == 6123

    def test_angles_atom_counts(self, filename):
        u = mda.Universe(filename)
        assert len(u.atoms[[0]].angles), 9
        assert len(u.atoms[[42]].angles), 2

    def test_angles_identity(self, top):
        vals = top.angles.values
        for b in ((1, 0, 2), (1, 0, 3), (1, 0, 4)):
            assert (b in vals) or (b[::-1] in vals)

    def test_dihedrals_total_counts(self, top):
        assert len(top.dihedrals.values) == 8921

    def test_dihedrals_atom_counts(self, filename):
        u = mda.Universe(filename)
        assert len(u.atoms[[0]].dihedrals) == 14

    def test_dihedrals_identity(self, top):
        vals = top.dihedrals.values
        for b in ((0, 4, 6, 7), (0, 4, 6, 8), (0, 4, 6, 9), (0, 4, 17, 18)):
            assert (b in vals) or (b[::-1] in vals)


class TestNAMDPSFParser(PSFBase):
    """Testfiles provided by JiyongPark77.

    NAMD/VMD XPLOR-style PSF file (using CGENFF residues/atoms).

    https://github.com/MDAnalysis/mdanalysis/issues/107
    """
    ref_filename = PSF_NAMD
    expected_n_atoms = 130
    expected_n_residues = 6
    expected_n_segments = 1


class TestPSFParser2(PSFBase):
    ref_filename = XYZ_psf
    expected_n_atoms = 1284
    expected_n_residues = 152
    expected_n_segments = 4

    def test_as_universe_resids(self):
        u = mda.Universe(XYZ_psf, XYZ)

        # each segment has [380, 381, 382, 383] as the first
        # 4 residues
        for seg in u.segments:
            assert_equal(seg.residues.resids[:4], [380, 381, 382, 383])

class TestPSFParserNoTop(PSFBase):
    ref_filename = PSF_notop
    expected_n_atoms = 3341
    expected_n_residues = 214
    expected_n_segments = 1

    def test_bonds_total_counts(self, top):
        assert len(top.bonds.values) == 0

    def test_angles_total_counts(self, top):
        assert len(top.angles.values) == 0

    def test_dihedrals_total_counts(self, top):
        assert len(top.dihedrals.values) == 0
    
    def test_impropers_total_counts(self, top):
        assert len(top.impropers.values) == 0

def test_psf_nosegid():
    """Issue #121"""
    u = mda.Universe(PSF_nosegid)
    assert isinstance(u, mda.Universe)
    assert u.atoms.n_atoms == 98
    assert_equal(u.segments.segids, ["SYSTEM"])



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import numpy as np

import MDAnalysis as mda
import pytest
from MDAnalysis.coordinates.memory import MemoryReader
from MDAnalysisTests.datafiles import DCD, PSF
from MDAnalysisTests.coordinates.base import (BaseReference,
                                              MultiframeReaderTest)
from MDAnalysis.coordinates.memory import Timestep
from numpy.testing import assert_equal, assert_almost_equal


class MemoryReference(BaseReference):
    def __init__(self):
        super(MemoryReference, self).__init__()

        self.topology = PSF
        self.trajectory = DCD
        self.universe = mda.Universe(PSF, DCD)

        self.n_atoms = self.universe.trajectory.n_atoms
        self.n_frames = self.universe.trajectory.n_frames

        self.dt = self.universe.trajectory.ts.dt
        self.dimensions = self.universe.trajectory.ts.dimensions
        self.totaltime = self.universe.trajectory.totaltime
        self.volume = self.universe.trajectory.ts.volume

        self.first_frame = Timestep(self.n_atoms)
        self.first_frame.positions = np.array(self.universe.trajectory[0])
        self.first_frame.frame = 0
        self.first_frame.time = self.first_frame.frame*self.dt

        self.second_frame = Timestep(self.n_atoms)
        self.second_frame.positions = np.array(self.universe.trajectory[1])
        self.second_frame.frame = 1
        self.second_frame.time = self.second_frame.frame*self.dt

        self.last_frame = Timestep(self.n_atoms)
        self.last_frame.positions = \
            np.array(self.universe.trajectory[self.n_frames - 1])
        self.last_frame.frame = self.n_frames - 1
        self.last_frame.time = self.last_frame.frame*self.dt

        self.jump_to_frame = self.first_frame.copy()
        self.jump_to_frame.positions = np.array(self.universe.trajectory[3])
        self.jump_to_frame.frame = 3
        self.jump_to_frame.time = self.jump_to_frame.frame*self.dt

    def reader(self, trajectory):
        return mda.Universe(self.topology,
                            trajectory, in_memory=True).trajectory

    def iter_ts(self, i):
        ts = self.universe.trajectory[i]
        # correct time because memory reader doesn't read the correct time
        ts.time = ts.frame * self.dt
        return ts


class TestMemoryReader(MultiframeReaderTest):
    @staticmethod
    @pytest.fixture(scope='class')
    def ref():
        return MemoryReference()

    def test_filename_transefer_to_memory(self):
        # MemoryReader should have a filename attribute set to the trajaectory filename
        universe = mda.Universe(PSF, DCD)
        universe.transfer_to_memory()
        assert_equal(universe.trajectory.filename, DCD)

    def test_filename_array(self):
        # filename attribute of MemoryReader should be None when generated from an array
        universe = mda.Universe(PSF, DCD)
        coordinates = universe.trajectory.timeseries(universe.atoms)
        universe2 = mda.Universe(PSF, coordinates, format=MemoryReader, order='afc')
        assert universe2.trajectory.filename is None

    def test_default_memory_layout(self):
        universe1 = mda.Universe(PSF, DCD, in_memory=True)
        universe2 = mda.Universe(PSF, DCD, in_memory=True, order='fac')
        assert_equal(universe1.trajectory.get_array().shape,
                     universe2.trajectory.get_array().shape)

    def test_iteration(self, ref, reader):
        frames = 0
        for i, frame in enumerate(reader):
            frames += 1
        assert frames == ref.n_frames

    def test_extract_array_afc(self,reader):
        assert_equal(reader.timeseries(order='afc').shape, (3341, 98, 3))

    def test_extract_array_afc(self, reader):
        assert_equal(reader.timeseries(order='afc').shape, (3341, 98, 3))

    def test_extract_array_fac(self, reader):
        assert_equal(reader.timeseries(order='fac').shape, (98, 3341, 3))

    def test_extract_array_cfa(self, reader):
        assert_equal(reader.timeseries(order='cfa').shape, (3, 98, 3341))

    def test_extract_array_acf(self, reader):
        assert_equal(reader.timeseries(order='acf').shape, (3341, 3, 98))

    def test_extract_array_fca(self, reader):
        assert_equal(reader.timeseries(order='fca').shape, (98, 3, 3341))

    def test_extract_array_caf(self, reader):
        assert_equal(reader.timeseries(order='caf').shape, (3, 3341, 98))

    def test_timeseries_skip1(self, ref, reader):
        assert_equal(reader.timeseries(ref.universe.atoms).shape,
                     (3341, 98, 3))

    def test_timeseries_skip10(self, reader):
        # Check that timeseries skip works similar to numpy slicing
        array1 = reader.timeseries(step=10)
        array2 = reader.timeseries()[:,::10,:]
        assert_equal(array1, array2)

    def test_timeseries_view(self, reader):
        # timeseries() is expected to provide a view of the underlying array
        assert reader.timeseries().base is reader.get_array()

    def test_timeseries_subarray_view(self, reader):
        # timeseries() is expected to provide a view of the underlying array
        # also in the case where we slice the array using the start, stop and
        # step options.
        assert reader.timeseries(start=5,stop=15,step=2,order='fac').base is\
               reader.get_array()

    def test_timeseries_view_from_universe_atoms(self, ref, reader):
        # timeseries() is expected to provide a view of the underlying array
        # also in the special case when asel=universe.atoms.
        selection = ref.universe.atoms
        assert reader.timeseries(asel=selection).base is reader.get_array()

    def test_timeseries_view_from_select_all(self, ref, reader):
        # timeseries() is expected to provide a view of the underlying array
        # also in the special case when using "all" in selections.
        selection = ref.universe.select_atoms("all")
        assert_equal(reader.timeseries(
            asel=selection).base is reader.get_array(),
            True)

    def test_timeseries_noview(self, ref, reader):
        # timeseries() is expected NOT to provide a view of the underlying array
        # for any other selection than "all".
        selection = ref.universe.select_atoms("name CA")
        assert reader.timeseries(asel=selection).base is not reader.get_array()

    def test_repr(self, reader):
        str_rep = str(reader)
        expected = "<MemoryReader with 98 frames of 3341 atoms>"
        assert_equal(str_rep, expected)

    def test_get_writer_1(self):
        pass

    def test_get_writer_2(self):
        pass

    def test_float32(self, ref):
        # Check that we get float32 positions even when initializing with float64
        coordinates = np.random.uniform(size=(100, ref.universe.atoms.n_atoms, 3)).cumsum(0)
        universe = mda.Universe(ref.universe.filename, coordinates, format=MemoryReader)
        assert_equal(universe.trajectory.get_array().dtype, np.dtype('float32'))

    def test_position_assignation(self, reader):
        # When coordinates are assigned to a timestep, is the change persistent?
        new_positions = np.ones_like(reader.ts.positions, dtype=np.float32)
        reader.ts.positions = new_positions
        reader[0]
        assert_almost_equal(reader.ts.positions, new_positions)


class TestMemoryReaderVelsForces(object):
    @staticmethod
    @pytest.fixture(params=['2d', '3d'])
    def ref_pos(request):
        if request.param == '2d':
            return np.arange(30).reshape(10, 3)
        elif request.param == '3d':
            return np.arange(30).reshape(1, 10, 3)

    @staticmethod
    @pytest.fixture(params=['2d', '3d'])
    def ref_vels(request):
        if request.param == '2d':
            return np.arange(30).reshape(10, 3) + 100
        elif request.param == '3d':
            return np.arange(30).reshape(1, 10, 3) + 100

    @staticmethod
    @pytest.fixture(params=['2d', '3d'])
    def ref_forces(request):
        if request.param == '2d':
            return np.arange(30).reshape(10, 3) + 1000
        elif request.param == '3d':
            return np.arange(30).reshape(1, 10, 3) + 1000

    @staticmethod
    def assert_equal_dims(arr1, arr2):
        if arr2.ndim == 3:
            assert_equal(arr1, arr2[0])
        elif arr2.ndim == 2:
            assert_equal(arr1, arr2)

    def test_velocities(self, ref_pos, ref_vels):
        mr = MemoryReader(ref_pos,
                          velocities=ref_vels)

        assert mr.ts.has_velocities
        self.assert_equal_dims(mr.ts.velocities, ref_vels)
        assert not mr.ts.has_forces

    def test_forces(self, ref_pos, ref_forces):
        mr = MemoryReader(ref_pos,
                          forces=ref_forces)

        assert not mr.ts.has_velocities
        assert mr.ts.has_forces
        self.assert_equal_dims(mr.ts.forces, ref_forces)

    def test_both(self, ref_pos, ref_vels, ref_forces):
        mr = MemoryReader(ref_pos,
                          velocities=ref_vels,
                          forces=ref_forces)
        assert mr.ts.has_velocities
        self.assert_equal_dims(mr.ts.velocities, ref_vels)
        assert mr.ts.has_forces
        self.assert_equal_dims(mr.ts.forces, ref_forces)

    @pytest.mark.parametrize('param', ['velocities', 'forces'])
    def test_wrongshape(self, ref_pos, param):
        with pytest.raises(ValueError):
            mr = MemoryReader(ref_pos, **{param: np.zeros((3, 2, 1))})


class TestDimensions(object):
    @staticmethod
    @pytest.fixture
    def ref_pos():
        return np.arange(270).reshape(3, 30, 3)

    @staticmethod
    @pytest.fixture
    def ref_box():
        return np.arange(18).reshape(3, 6)

    def test_single_box(self, ref_pos):
        box = np.array([3, 4, 5, 90, 90, 90])

        mr = MemoryReader(ref_pos, dimensions=box)

        for ts in mr:
            assert_equal(ts.dimensions, box)

    def test_varying_box(self, ref_pos, ref_box):
        mr = MemoryReader(ref_pos, dimensions=ref_box)

        for i, ts in enumerate(mr):
            assert_equal(ts.dimensions, ref_box[i])

    def test_wrong_length(self, ref_pos):
        bad_box = np.arange(12).reshape(2, 6)

        with pytest.raises(ValueError):
            mr = MemoryReader(ref_pos, dimensions=bad_box)

    def test_wrong_shape(self, ref_pos):
        bad_box = np.arange(15).reshape(3, 5)

        with pytest.raises(ValueError):
            mr = MemoryReader(ref_pos, dimensions=bad_box)


class TestMemoryReaderModifications(object):
    # check that modifying MR things behaves as expected
    # in general, modifying the Timestep should be *permanent*
    # this is unlike other Readers!
    n_atoms = 10
    n_frames = 4

    @pytest.fixture()
    def mr_reader(self):
        pos = np.arange(self.n_frames * self.n_atoms * 3).reshape(
            self.n_frames, self.n_atoms, 3)
        vel = np.arange(self.n_frames * self.n_atoms * 3).reshape(
            self.n_frames, self.n_atoms, 3) + 200
        frc = np.arange(self.n_frames * self.n_atoms * 3).reshape(
            self.n_frames, self.n_atoms, 3) + 400
        box = np.arange(self.n_frames * 6).reshape(self.n_frames, 6) + 600

        return MemoryReader(pos,
                            velocities=vel,
                            forces=frc,
                            dimensions=box)

    @pytest.fixture()
    def mr_universe(self, mr_reader):
        u = mda.Universe.empty(self.n_atoms)
        u.trajectory = mr_reader

        return u

    @pytest.mark.parametrize('attr', ['positions', 'velocities', 'forces', 'dimensions'])
    def test_copying(self, mr_reader, attr):
        mr2 = mr_reader.copy()
        # update the attribute
        ts = mr2.ts
        setattr(ts, attr, 7)
        # check the change worked
        assert_almost_equal(getattr(ts, attr), 7)
        assert ts.positions.shape == (self.n_atoms, 3)
        assert ts.velocities.shape == (self.n_atoms, 3)
        assert ts.forces.shape == (self.n_atoms, 3)
        assert ts.dimensions.shape == (6,)
        # move the Reader around, forcing updates of ts
        ts = mr2[2]
        ts = mr2[0]
        # check our old change is still there
        assert_almost_equal(getattr(ts, attr), 7)

    @pytest.mark.parametrize('attr', ['positions', 'velocities', 'forces', 'dimensions'])
    def test_attr_set(self, mr_universe, attr):
        # same as above, but via a Universe/AtomGroup
        u = mr_universe
        ts = u.trajectory[0]

        setattr(ts, attr, 7)

        assert_almost_equal(getattr(ts, attr), 7)

        ts = u.trajectory[2]
        ts = u.trajectory[0]

        assert_almost_equal(getattr(ts, attr), 7)
        assert u.atoms.positions.shape == (self.n_atoms, 3)
        assert u.atoms.velocities.shape == (self.n_atoms, 3)
        assert u.atoms.forces.shape == (self.n_atoms, 3)
        assert u.atoms.dimensions.shape == (6,)

    @pytest.mark.parametrize('attr', ['velocities', 'forces', 'dimensions'])
    def test_non_numpy_arr(self, attr):
        with pytest.raises(TypeError):
            mr = MemoryReader(np.zeros((10, 30, 3)),
                              **{attr: 'not an array'})



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest

import os
from numpy.testing import (
    assert_equal, assert_array_equal,
    assert_array_almost_equal, TestCase,
    assert_almost_equal
)

from MDAnalysisTests.datafiles import (
    mol2_molecules, mol2_molecule, mol2_broken_molecule,
    mol2_zinc, mol2_comments_header, mol2_ligand, mol2_sodium_ion
)
from MDAnalysis import Universe
import MDAnalysis as mda
from MDAnalysisTests import make_Universe


class TestMol2(object):
    def setup_method(self):
        self.outfile = 'test.mol2'

    def test_read(self):
        u = Universe(mol2_molecules)
        assert_equal(len(u.atoms), 49)
        assert_equal(u.trajectory.n_frames, 200)

        u.trajectory[199]
        assert_array_almost_equal(u.atoms.positions[0], [1.7240, 11.2730, 14.1200])

    def test_read_statusbit(self):
        u = Universe(mol2_ligand)
        assert_equal(len(u.atoms), 297)
        assert_equal(u.trajectory.n_frames, 1)

    def test_write(self, tmpdir):
        ref = Universe(mol2_molecules)
        with tmpdir.as_cwd():
            ref.atoms.write(self.outfile)
            u = Universe(self.outfile)
            assert_equal(len(u.atoms), len(ref.atoms))
            assert_equal(len(u.trajectory), 1)
            assert_array_equal(u.atoms.positions, ref.atoms.positions)

    def test_write_selection(self, tmpdir):
        ref = Universe(mol2_molecule)
        gr0 = ref.select_atoms("name C*")
        with tmpdir.as_cwd():
            gr0.write(self.outfile)
            u = Universe(self.outfile)
            gr1 = u.select_atoms("name C*")
            assert_equal(len(gr0), len(gr1))

    def test_write_in_loop(self, tmpdir):
        ref = Universe(mol2_molecules)

        with tmpdir.as_cwd():
            with mda.Writer(self.outfile) as W:
                for ts in ref.trajectory:
                    W.write(ref.atoms)
            u = Universe(self.outfile)
            assert_equal(len(u.atoms), len(ref.atoms))
            assert_equal(len(u.trajectory), len(ref.trajectory))
            assert_array_equal(u.atoms.positions, ref.atoms.positions)
            u.trajectory[199]
            ref.trajectory[199]
            assert_array_equal(u.atoms.positions, ref.atoms.positions)

    def test_broken_molecule(self):
        with pytest.raises(ValueError):
            Universe(mol2_broken_molecule)

        # This doesn't work with 2.6
        # Checks the text of the error message, so it low priority
        #with self.assertRaises(Exception) as context:
        #    u = Universe(mol2_broken_molecule)
        #self.assertEqual("The mol2 block (BrokenMolecule.mol2:0) has no atoms" in context.exception.message,
        # True)

    def test_comments_header(self):
        u = Universe(mol2_comments_header)
        assert_equal(len(u.atoms), 9)
        assert_equal(u.trajectory.n_frames, 2)
        u.trajectory[1]
        assert_array_almost_equal(u.atoms.positions[2], [-12.2710, -1.9540, -16.0480])

    def test_no_bonds(self, tmpdir):
        # Issue #3057
        u = Universe(mol2_sodium_ion)
        ag = u.atoms
        assert not hasattr(ag, "bonds")
        with tmpdir.as_cwd():
            outfile = 'test.mol2'
            ag.write(outfile)
            u2 = Universe(outfile)
        assert not hasattr(u2.atoms, "bonds")


class TestMol2_traj(TestCase):
    def setUp(self):
        self.universe = Universe(mol2_molecules)
        self.traj = self.universe.trajectory
        self.ts = self.universe.coord

    def tearDown(self):
        del self.universe
        del self.traj
        del self.ts

    def test_rewind_traj(self):
        self.traj.rewind()
        assert_equal(self.ts.frame, 0, "rewinding to frame 0")

    def test_next_traj(self):
        self.traj.rewind()
        self.traj.next()
        assert_equal(self.ts.frame, 1, "loading frame 1")

    def test_jump_traj(self):
        self.traj[15]  # index is 0-based and frames are 0-based
        assert_equal(self.ts.frame, 15, "jumping to frame 15")

    def test_jump_lastframe_traj(self):
        self.traj[-1]
        assert_equal(self.ts.frame, 199, "indexing last frame with traj[-1]")

    def test_slice_traj(self):
        frames = [ts.frame for ts in self.traj[5:17:3]]
        assert_equal(frames, [5, 8, 11, 14], "slicing traj [5:17:3]")

    def test_reverse_traj(self):
        frames = [ts.frame for ts in self.traj[20:5:-1]]
        assert_equal(frames, list(range(20, 5, -1)),
                     "reversing traj [20:5:-1]")

    def test_n_frames(self):
        assert_equal(self.universe.trajectory.n_frames, 200, "wrong number of frames in traj")


class TestMOL2NoSubstructure(object):
    """MOL2 file without substructure

    """
    n_atoms = 45

    def test_load(self):
        r = mda.coordinates.MOL2.MOL2Reader(mol2_zinc, n_atoms=self.n_atoms)
        assert r.n_atoms == 45

    def test_universe(self):
        u = mda.Universe(mol2_zinc)
        assert len(u.atoms) == self.n_atoms

    def test_write_nostructure(self, tmpdir):
        with tmpdir.as_cwd():
            outfile = 'test.mol2'

            u = mda.Universe(mol2_zinc)
            with mda.Writer(outfile) as W:
                W.write(u.atoms)

            u2 = mda.Universe(outfile)

            assert len(u.atoms) == len(u2.atoms)


def test_mol2_write_NIE(tmpdir):
    with tmpdir.as_cwd():
        outfile = os.path.join('test.mol2')
        u = make_Universe(trajectory=True)
        with pytest.raises(NotImplementedError):
            u.atoms.write(outfile)

def test_mol2_multi_write(tmpdir):
    # see: gh-2678
    with tmpdir.as_cwd():
        u = mda.Universe(mol2_molecules)
        u.atoms[:4].write('group1.mol2')
        u.atoms[:4].write('group1.mol2')


def test_mol2_universe_write(tmpdir):
    # see Issue 2717
    with tmpdir.as_cwd():
        outfile = 'test.mol2'

        u = mda.Universe(mol2_comments_header)

        with mda.Writer(outfile) as W:
            W.write(u)

        u2 = mda.Universe(outfile)

        assert_almost_equal(u.atoms.positions, u2.atoms.positions)
        # MDA does not current implement @<TRIPOS>CRYSIN reading
        assert u2.dimensions is None



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
from __future__ import absolute_import

import numpy as np
from numpy.testing import assert_equal, assert_almost_equal
import pytest

import MDAnalysis as mda
from MDAnalysisTests.datafiles import (PSF, DCD, mol2_comments_header, XYZ_mini,
                                       BATArray, TPR, XTC)
from MDAnalysis.analysis.bat import BAT


class TestBAT(object):
    @pytest.fixture()
    def selected_residues(self):
        u = mda.Universe(mol2_comments_header)
        ag = u.select_atoms("all")
        return ag

    @pytest.fixture()
    def bat(self, selected_residues):
        R = BAT(selected_residues)
        R.run()
        return R.results.bat

    @pytest.fixture
    def bat_npz(self, tmpdir, selected_residues):
        filename = str(tmpdir / 'test_bat_IO.npy')
        R = BAT(selected_residues)
        R.run()
        R.save(filename)
        return filename

    def test_bat_root_selection(self, selected_residues):
        R = BAT(selected_residues)
        assert_equal(R._root.indices, [8, 2, 1],
                     err_msg="error: incorrect root atoms selected")

    def test_bat_number_of_frames(self, bat):
        assert_equal(len(bat),
                     2,
                     err_msg="error: list is not length of trajectory")

    def test_bat_coordinates(self, bat):
        test_bat = np.load(BATArray)
        assert_almost_equal(
            bat,
            test_bat,
            5,
            err_msg="error: BAT coordinates should match test values")

    def test_bat_coordinates_single_frame(self, selected_residues):
        bat = BAT(selected_residues).run(start=1, stop=2).results.bat
        test_bat = [np.load(BATArray)[1]]
        assert_almost_equal(
            bat,
            test_bat,
            5,
            err_msg="error: BAT coordinates should match test values")

    def test_bat_reconstruction(self, selected_residues, bat):
        R = BAT(selected_residues)
        XYZ = R.Cartesian(bat[0])
        assert_almost_equal(XYZ, selected_residues.positions, 5,
            err_msg="error: Reconstructed Cartesian coordinates " + \
                    "don't match original")

    def test_bat_IO(self, bat_npz, selected_residues, bat):
        R2 = BAT(selected_residues, filename=bat_npz)
        test_bat = R2.results.bat
        assert_almost_equal(
            bat,
            test_bat,
            5,
            err_msg="error: Loaded BAT coordinates should match test values")

    def test_bat_nobonds(self):
        u = mda.Universe(XYZ_mini)
        errmsg = "AtomGroup has no attribute bonds"
        with pytest.raises(AttributeError, match=errmsg):
            Z = BAT(u.atoms)

    def test_bat_bad_initial_atom(self, selected_residues):
        errmsg = 'Initial atom is not a terminal atom'
        with pytest.raises(ValueError, match=errmsg):
            R = BAT(selected_residues, initial_atom = selected_residues[0])

    def test_bat_disconnected_atom_group(self):
        u = mda.Universe(PSF, DCD)
        selected_residues = u.select_atoms("resid 1-3") + \
            u.select_atoms("resid 5-7")
        errmsg = 'Additional torsions not found.'
        with pytest.raises(ValueError, match=errmsg):
            R = BAT(selected_residues)

    def test_bat_multifragments_atomgroup(self):
        u = mda.Universe(TPR, XTC)
        errmsg = 'AtomGroup has more than one molecule'
        with pytest.raises(ValueError, match=errmsg):
            BAT(u.select_atoms('resname SOL'))

    def test_bat_incorrect_dims(self, bat_npz):
        u = mda.Universe(PSF, DCD)
        selected_residues = u.select_atoms("resid 1-3")
        errmsg = 'Dimensions of array in loaded file'
        with pytest.raises(ValueError, match=errmsg):
            R = BAT(selected_residues, filename=bat_npz)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import numpy as np
from numpy.testing import (
    assert_equal,
)
import pytest

import MDAnalysis as mda
from MDAnalysis.core.topologyattrs import Bonds
from MDAnalysis.core import groups
from MDAnalysis import NoDataError

from MDAnalysisTests import make_Universe
from MDAnalysisTests.datafiles import TPR, XTC


# Also used in topology/test_guessers
def make_starshape():
    u = make_Universe()
    bonds = []
    for seg in range(5):
        segbase = seg * 25
        for res in range(5):
            # offset for atoms in this res
            base = segbase + 5 * res
            bonds.append((0 + base, 1 + base))
            bonds.append((1 + base, 2 + base))
            bonds.append((1 + base, 3 + base))
            bonds.append((1 + base, 4 + base))
            if not res == 4:  # last res doesn't link onwards
                bonds.append((4 + base, 5 + base))
    u.add_TopologyAttr(Bonds(bonds))
    return u


def case1():
    return make_starshape()


def case2():
    u = make_Universe()
    bonds = []
    for seg in range(5):
        segbase = seg * 25
        for res in range(5):
            # offset for atoms in this res
            base = segbase + 5 * res
            bonds.append((0 + base, 1 + base))
            bonds.append((1 + base, 2 + base))
            bonds.append((2 + base, 3 + base))
            bonds.append((3 + base, 4 + base))
            bonds.append((1 + base, 4 + base))
            if not res == 4:  # last res doesn't link onwards
                bonds.append((0 + base, 5 + base))
    u.add_TopologyAttr(Bonds(bonds))
    return u


class TestFragments(object):
    r"""Use 125 atom test Universe

    5 segments of 5 residues of 5 atoms


    Case1
    -----

    Star shapes to try and test the branching prediction

      o   |   o   |   o
      |   |   |   |   |
    o-o-o-|-o-o-o-|-o-o-o
      |   |   |   |   |
      o   |   o   |x3 o


    Case2
    -----

    4-ring pendants to test cyclic conditions

      o------o------o
      |      |      |
      o      o      o
     / \    / \    / \
    o   o  o   o  o   o
     \ /    \ /    \ /
      o      o      o
    Test ring molecules?
    """

    @pytest.mark.parametrize('u', (
            case1(),
            case2()
    ))
    def test_total_frags(self, u):
        fragments = u.atoms.fragments
        fragindices = u.atoms.fragindices
        # should be 5 fragments of 25 atoms
        assert len(fragments) == 5
        for frag in fragments:
            assert len(frag) == 25
        # number of fragindices must correspond to number of atoms:
        assert len(fragindices) == len(u.atoms)
        # number of unique fragindices must correspond to number of fragments:
        assert len(np.unique(fragindices)) == len(fragments)
        # check fragindices dtype:
        assert fragindices.dtype == np.intp
        #check n_fragments
        assert u.atoms.n_fragments == len(fragments)

    @pytest.mark.parametrize('u', (
            case1(),
            case2()
    ))
    def test_frag_external_ordering(self, u):
        # check fragments and fragindices are sorted correctly:
        for i, frag in enumerate(u.atoms.fragments):
            assert frag[0].index == i * 25
            assert np.unique(frag.fragindices)[0] == i

    @pytest.mark.parametrize('u', (
            case1(),
            case2()
    ))
    def test_frag_internal_ordering(self, u):
        # check atoms are sorted within fragments and have the same fragindex:
        for i, frag in enumerate(u.atoms.fragments):
            assert_equal(frag.ix, np.arange(25) + i * 25)
            assert len(np.unique(frag.fragindices)) == 1
            assert frag.n_fragments == 1

    @pytest.mark.parametrize('u', (
            case1(),
            case2()
    ))
    def test_atom_access(self, u):
        # check atom can access fragment and fragindex:
        for at in (u.atoms[0], u.atoms[76], u.atoms[111]):
            frag = at.fragment
            assert isinstance(frag, groups.AtomGroup)
            assert len(frag) == 25
            assert at in frag
            fragindex = at.fragindex
            assert isinstance(fragindex, int)
            with pytest.raises(AttributeError):
                x = at.n_fragments

    @pytest.mark.parametrize('u', (
            case1(),
            case2()
    ))
    def test_atomgroup_access(self, u):
        # check atomgroup can access fragments
        # first 60 atoms have 3 fragments, given as tuple
        # each fragment should still be 25 atoms
        ag = u.atoms[:60]
        frags = ag.fragments
        assert len(frags) == 3
        assert isinstance(frags, tuple)
        for frag in frags:
            assert len(frag) == 25
        # same for fragindices:
        fragindices = ag.fragindices
        assert len(fragindices) == 60
        assert len(np.unique(fragindices)) == 3
        assert ag.n_fragments == 3

    def test_empty_atomgroup_access(self):
        ag = mda.AtomGroup([], case1())
        assert ag.fragments == tuple()
        assert_equal(ag.fragindices, np.array([], dtype=np.int64))
        assert ag.n_fragments == 0

    def test_atomgroup_fragments_nobonds_NDE(self):
        # should raise NDE
        u = make_Universe()
        ag = u.atoms[:10]
        with pytest.raises(NoDataError):
            getattr(ag, 'fragments')
        with pytest.raises(NoDataError):
            getattr(ag, 'fragindices')
        with pytest.raises(NoDataError):
            getattr(ag, 'n_fragments')

    def test_atom_fragment_nobonds_NDE(self):
        # should raise NDE
        u = make_Universe()
        with pytest.raises(NoDataError):
            getattr(u.atoms[10], 'fragment')
        with pytest.raises(NoDataError):
            getattr(u.atoms[10], 'fragindex')

    def test_atomgroup_fragment_cache_invalidation_bond_making(self):
        u = case1()
        fgs = u.atoms.fragments
        assert fgs is u.atoms._cache['fragments']
        assert u.atoms._cache_key in u._cache['_valid']['fragments']
        u.add_bonds((fgs[0][-1] + fgs[1][0],))  # should trigger invalidation
        assert 'fragments' not in u._cache['_valid']
        assert len(fgs) > len(u.atoms.fragments)  # recomputed

    def test_atomgroup_fragment_cache_invalidation_bond_breaking(self):
        u = case1()
        fgs = u.atoms.fragments
        assert fgs is u.atoms._cache['fragments']
        assert u.atoms._cache_key in u._cache['_valid']['fragments']
        u.delete_bonds((u.atoms.bonds[3],))  # should trigger invalidation
        assert 'fragments' not in u._cache['_valid']
        assert len(fgs) < len(u.atoms.fragments)  # recomputed


def test_tpr_fragments():
    ag = mda.Universe(TPR, XTC).atoms
    frags = ag.fragments
    fragindices = ag.fragindices

    assert len(frags[0]) == 3341
    assert len(fragindices) == len(ag)
    assert len(np.unique(fragindices)) == len(frags)
    assert ag.n_fragments == len(frags)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import numpy as np
from numpy.testing import assert_almost_equal
import MDAnalysis
from MDAnalysis.visualization import (streamlines,
                                      streamlines_3D)
from MDAnalysis.coordinates.XTC import XTCWriter
from MDAnalysisTests.datafiles import Martini_membrane_gro
import pytest
import matplotlib.pyplot as plt
import os

@pytest.fixture(scope="session")
def univ():
    u = MDAnalysis.Universe(Martini_membrane_gro)
    return u

@pytest.fixture(scope="session")
def membrane_xtc(tmpdir_factory, univ):
    x_delta, y_delta, z_delta  = 0.5, 0.3, 0.2
    tmp_xtc = tmpdir_factory.mktemp('streamlines').join('dummy.xtc')

    with XTCWriter(str(tmp_xtc), n_atoms=univ.atoms.n_atoms) as xtc_writer:
        for i in range(5):
           univ.atoms.translate([x_delta, y_delta, z_delta])
           xtc_writer.write(univ.atoms)
           x_delta += 0.1
           y_delta += 0.08
           z_delta += 0.02
    return str(tmp_xtc)

def test_streamplot_2D(membrane_xtc, univ, tmpdir):
    # simple roundtrip test to ensure that
    # a plot is generated by the 2D streamplot
    # code
    u1, v1, avg, std = streamlines.generate_streamlines(topology_file_path=Martini_membrane_gro,
                                                        trajectory_file_path=membrane_xtc,
                                                        grid_spacing=20,
                                                        MDA_selection='name PO4',
                                                        start_frame=1,
                                                        end_frame=2,
                                                        xmin=univ.atoms.positions[...,0].min(),
                                                        xmax=univ.atoms.positions[...,0].max(),
                                                        ymin=univ.atoms.positions[...,1].min(),
                                                        ymax=univ.atoms.positions[...,1].max(),
                                                        maximum_delta_magnitude=2.0,
                                                        num_cores=1)
    x = np.linspace(univ.atoms.positions[...,0].min(),
                    univ.atoms.positions[...,0].max(),
                    5)
    y = np.linspace(univ.atoms.positions[...,1].min(),
                    univ.atoms.positions[...,1].max(),
                    5)
    speed = np.sqrt(u1*u1 + v1*v1)
    fig = plt.figure()
    ax = fig.add_subplot(111, aspect='equal')
    ax.streamplot(x, y, u1, v1, density=(10,10), color=speed, linewidth=3*speed/speed.max(),
                  cmap=plt.cm.viridis)
    plot_outpath = os.path.join(str(tmpdir), 'testing_streamline.png')
    fig.savefig(plot_outpath, dpi=300)

    with open(plot_outpath, 'rb'):
        pass

def test_streamplot_3D(membrane_xtc, univ, tmpdir):
    # because mayavi is too heavy of a dependency
    # for a roundtrip plotting test, simply
    # aim to check for sensible values
    # returned by generate_streamlines_3d
    dx, dy, dz = streamlines_3D.generate_streamlines_3d(topology_file_path=Martini_membrane_gro,
                                                        trajectory_file_path=membrane_xtc,
                                                        grid_spacing=20,
                                                        MDA_selection='name PO4',
                                                        start_frame=1,
                                                        end_frame=2,
                                                        xmin=univ.atoms.positions[...,0].min(),
                                                        xmax=univ.atoms.positions[...,0].max(),
                                                        ymin=univ.atoms.positions[...,1].min(),
                                                        ymax=univ.atoms.positions[...,1].max(),
                                                        zmin=univ.atoms.positions[...,2].min(),
                                                        zmax=univ.atoms.positions[...,2].max(),
                                                        maximum_delta_magnitude=2.0,
                                                        num_cores=1)
    assert dx.shape == (5, 5, 2)
    assert dy.shape == (5, 5, 2)
    assert dz.shape == (5, 5, 2)
    assert_almost_equal(dx[4, 4, 0], 0.700004, decimal=5)
    assert_almost_equal(dy[0, 0, 0], 0.460000, decimal=5)
    assert_almost_equal(dz[2, 2, 0], 0.240005, decimal=5)
